__all__ = ('ChannelMetadataPrivateGroup',)

from scarletio import copy_docs

from ...bases import ICON_TYPE_NONE, Slotted
from ...permission.permission import PERMISSION_GROUP, PERMISSION_GROUP_OWNER, PERMISSION_NONE

from .fields import parse_name, parse_owner_id, put_name_into, put_owner_id_into, validate_name, validate_owner_id

from .base import CHANNEL_METADATA_ICON_SLOT
from .private_base import ChannelMetadataPrivateBase


class ChannelMetadataPrivateGroup(ChannelMetadataPrivateBase, metaclass=Slotted):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    users : `list` of ``ClientUserBase``
        The users in the channel.
    icon_hash : `int`
        The channel's icon's hash in `uint128`.
    icon_type : ``iconType``
        The channel's icon's type.
    name : `str`
        The channel's display name. Can be empty string if the channel has no name.
    owner_id : `int`
        The group channel's owner's id.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('name', 'owner_id')
    
    icon = CHANNEL_METADATA_ICON_SLOT
    
    @copy_docs(ChannelMetadataPrivateBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataPrivateBase.__hash__(self)
        
        # name
        name = self.name
        if name:
            hash_value ^= hash(name)
        
        # owner_id
        hash_value ^= self.owner_id
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataPrivateBase.__new__)
    def __new__(cls, keyword_parameters):
        self = ChannelMetadataPrivateBase.__new__(cls, keyword_parameters)
        
        # icon
        self.icon = cls.icon.parse_from_keyword_parameters(keyword_parameters, allow_data = True)
        
        return self
    
    
    @copy_docs(ChannelMetadataPrivateBase._created)
    def _created(self, channel_entity, client):
        if (client is not None):
            client.group_channels[channel_entity.id] = channel_entity
    
        
    @copy_docs(ChannelMetadataPrivateBase._delete)
    def _delete(self, channel_entity, client):
        if (client is not None):
            try:
                del client.group_channels[channel_entity.id]
            except KeyError:
                pass
    
    
    @copy_docs(ChannelMetadataPrivateBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataPrivateBase._is_equal_same_type(self, other):
            return False
        
        # icon
        if self.icon_hash != other.icon_hash:
            return False
        
        if self.icon_type is not other.icon_type:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # owner_id
        if self.owner_id != other.owner_id:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataPrivateBase._get_processed_name)
    def _get_processed_name(self):
        name = self.name
        if name:
            return name
        
        users = self.users
        if users:
            return ', '.join([user.name for user in users])
        
        return 'Unnamed'
    
    
    @copy_docs(ChannelMetadataPrivateBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataPrivateBase._update_attributes(self, data)
        
        # icon
        self._set_icon(data)
        
        # name
        self.name = parse_name(data)
        
        # owner_id
        self.owner_id = parse_owner_id(data)
    
    
    @copy_docs(ChannelMetadataPrivateBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataPrivateBase._difference_update_attributes(self, data)
        
        # icon
        self._update_icon(data, old_attributes)
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # owner_id
        owner_id = parse_owner_id(data)
        if self.owner_id != owner_id:
            old_attributes['owner_id'] = self.owner_id
            self.owner_id = owner_id
        
        return old_attributes

    
    @copy_docs(ChannelMetadataPrivateBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        if self.owner_id == user.id:
            return PERMISSION_GROUP_OWNER
        elif user in self.users:
            return PERMISSION_GROUP
        else:
            return PERMISSION_NONE
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._from_partial_data)
    def _from_partial_data(cls, data):
        self = super(ChannelMetadataPrivateGroup, cls)._from_partial_data(data)
        
        if (data is not None):
            name = data.get('name', None)
            if (name is not None):
                self.name = name
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataPrivateGroup, cls)._create_empty()
        
        self.name = ''
        self.owner_id = 0
        
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase.precreate)
    def precreate(cls, keyword_parameters):
        self = super(ChannelMetadataPrivateGroup, cls).precreate(keyword_parameters)
        
        # icon
        self.icon = cls.icon.parse_from_keyword_parameters(keyword_parameters)
        
        return self
    
    
    @copy_docs(ChannelMetadataPrivateBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataPrivateBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            self.name = validate_name(name)
        
        # owner_id
        try:
            owner_id = keyword_parameters.pop('owner_id')
        except KeyError:
            pass
        else:
            self.owner_id = validate_owner_id(owner_id)
    
    
    @copy_docs(ChannelMetadataPrivateBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataPrivateBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # name
        put_name_into(self.name, data, defaults)
        
        # owner_id
        if include_internals:
            put_owner_id_into(self.owner_id, data, defaults)
        
        # icon
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        
        return data



__all__ = ()

from scarletio import Compound, Theory

from ...core import MESSAGES

from ...http import DiscordHTTPClient

from ...user import ClientUserBase, User
from ...utils import log_time_converter

from ..request_helpers import get_channel_id_and_message_id, get_emoji_from_reaction, get_reaction, get_user_id


class ClientCompoundReactionEndpoints(Compound):
    
    id: int
    http : DiscordHTTPClient
    
    @Theory
    async def message_get(self, message, *positional_parameters, force_update=False): ...
    
    
    async def reaction_add(self, message, emoji):
        """
        Adds a reaction on the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message on which the reaction will be put on.
        emoji : ``Emoji``, `str`
            The emoji to react with.
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
            - If `emoji`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji_as_reaction = get_reaction(emoji)
        
        await self.http.reaction_add(channel_id, message_id, emoji_as_reaction)
    
    
    async def reaction_delete(self, message, emoji, user):
        """
        Removes the specified reaction of the user from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reaction will be removed.
        emoji : ``Emoji``, `str`
            The emoji to remove.
        user : ```ClientUserBase``, `int`
            The user, who's reaction will be removed.
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
            - If `user` was not given neither as ``ClientUserBase`` nor `int`.
            - If `emoji` was not given as ``Emoji``, `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji_as_reaction = get_reaction(emoji)
        user_id = get_user_id(user)
        
        if user_id == self.id:
            await self.http.reaction_delete_own(channel_id, message_id, emoji_as_reaction)
        else:
            await self.http.reaction_delete(channel_id, message_id, emoji_as_reaction, user_id)
    
    
    async def reaction_delete_emoji(self, message, emoji):
        """
        Removes all the reaction of the specified emoji from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reactions will be removed.
        emoji : ``Emoji``, `str`
            The reaction to remove.
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
            - If `emoji`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        as_reaction = get_reaction(emoji)
        
        await self.http.reaction_delete_emoji(channel_id, message_id, as_reaction)
    
    
    async def reaction_delete_own(self, message, emoji):
        """
        Removes the specified reaction of the client from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reaction will be removed.
        emoji : ``Emoji``, `str`
            The emoji to remove.
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
            - If `emoji`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        as_reaction = get_reaction(emoji)
        await self.http.reaction_delete_own(channel_id, message_id, as_reaction)
    
    
    async def reaction_clear(self, message):
        """
        Removes all the reactions from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reactions will be cleared.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        
        await self.http.reaction_clear(channel_id, message_id)
    
    
    async def reaction_user_get_chunk(self, message, emoji, *, limit=None, after=None):
        """
        Requests the users, who reacted on the given message with the given emoji.
        
        If the message has no reactors at all or no reactors with that emoji, returns an empty list. If we know the
        emoji's every reactors we query the parameters from that.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message, what's reactions will be requested.
        emoji : ``Emoji``, `str`
            The emoji, what's reactors will be requested.
        limit : `None`, `int` = `None`, Optional (Keyword only)
            The amount of users to request. Can be in range [1:100]. Defaults to 25 by Discord.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the message's reactors were created.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
            - If `after` was passed with an unexpected type.
            - If `emoji`'s type is incorrect.
        ValueError
            The given `emoji` is not a valid reaction.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given neither as `None`, `int`.
            - If `limit` is out of the expected range [1:100].
        
        Notes
        -----
        `before` parameter is not supported by Discord.
        """
        if limit is None:
            limit = 25
        else:
            if __debug__:
                if not isinstance(limit, int):
                    raise AssertionError(
                        f'`limit` can be `None`, `int`, got {limit.__class__.__name__}; {limit!r}.'
                    )
                
                if limit < 1 or limit > 100:
                    raise AssertionError(
                        f'`limit` can be between in range [1:100], got {limit!r}.'
                    )
        
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji = get_emoji_from_reaction(emoji)
        
        try:
            message = MESSAGES[message_id]
        except KeyError:
            reactions = None
        else:
            reactions = message.reactions
            
            if reactions is None:
                return []
            
            try:
                line = reactions[emoji]
            except KeyError:
                return []
            
            if not line.unknown:
                after = 0 if after is None else log_time_converter(after)
                # before = 9223372036854775807 if before is None else log_time_converter(before)
                users = line.filter_after(limit, after)
                return users
        
        data = {}
        if limit != 25:
            data['limit'] = limit
        
        if (after is not None):
            data['after'] = log_time_converter(after)
        
        # if (before is not None):
        #     data['before'] = log_time_converter(before)
        
        data = await self.http.reaction_user_get_chunk(channel_id, message_id, emoji.as_reaction, data)
        
        users = [User.from_data(user_data) for user_data in data]
        
        if (reactions is not None):
            reactions._update_some_users(emoji, users)
        
        return users
    
    
    async def reaction_user_get_all(self, message, emoji):
        """
        Requests the all the users, which reacted on the message with the given message.
        
        If the message has no reactors at all or no reactors with that emoji returns an empty list. If the emoji's
        every reactors are known, then do requests are done.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message, what's reactions will be requested.
        emoji : ``Emoji``, `str`
            The emoji, what's reactors will be requested.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
            - If `emoji`'s type is incorrect.
        ValueError
            The given `emoji` is not a valid reaction.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji = get_emoji_from_reaction(emoji)
        
        try:
            message = MESSAGES[message_id]
        except KeyError:
            reactions = None
        else:
            reactions = message.reactions
            
            if (reactions is None):
                return []
            
            try:
                line = reactions[emoji]
            except KeyError:
                return []
            
            if not line.unknown:
                users = list(line)
                return users
        
        data = {'limit': 100, 'after': 0}
        users = []
        reaction = emoji.as_reaction
        
        while True:
            user_datas = await self.http.reaction_user_get_chunk(channel_id, message_id, reaction, data)
            users.extend(User.from_data(user_data) for user_data in user_datas)
            
            if len(user_datas) < 100:
                break
            
            data['after'] = users[-1].id
        
        if (reactions is not None):
            reactions._update_all_users(emoji, users)
        
        return users
    
    
    async def reaction_get_all(self, message):
        """
        Requests all the reactors for every emoji on the given message.
        
        Like the other reaction getting methods, this method prefers using the internal cache as well over doing a
        request.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message, what's reactions will be requested.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        message = MESSAGES.get(message_id, None)
        if (message is None):
            message = await self.message_get((channel_id, message_id))
        
        reactions = message.reactions
        if (reactions is not None) and reactions:
            users = []
            data = {'limit': 100}
            
            for emoji, line in reactions.items():
                if not line.unknown:
                    continue
                
                reaction = emoji.as_reaction
                data['after'] = 0
                
                while True:
                    user_datas = await self.http.reaction_user_get_chunk(channel_id, message_id, reaction, data)
                    users.extend(User.from_data(user_data) for user_data in user_datas)
                    
                    if len(user_datas) < 100:
                        break
                    
                    data['after'] = users[-1].id
                
                reactions._update_all_users(emoji, users)
                users.clear()
        
        return message



"""
Lavalink extension for hata.

Named after okuu's theme:

```
Solar Sect of Mystic Wisdom ~ Nuclear Fusion
```
"""

from .client import *
from .constants import *
from .event_handler_plugin import *
from .event_handlers import *
from .event_types import *
from .exceptions import *
from .filters import *
from .node import *
from .parsers import *
from .player import *
from .player_base import *
from .route_planner import *
from .stats import *
from .track import *
from .track_end_reasons import *

from . import  track_end_reasons as TRACK_END_REASONS

__all__ = (
    'TRACK_END_REASONS',
    *client.__all__,
    *constants.__all__,
    *event_handler_plugin.__all__,
    *event_handlers.__all__,
    *event_types.__all__,
    *exceptions.__all__,
    *filters.__all__,
    *node.__all__,
    *parsers.__all__,
    *player.__all__,
    *player_base.__all__,
    *route_planner.__all__,
    *stats.__all__,
    *track.__all__,
    *track_end_reasons.__all__,
)

from .. import register_library_extension, register_setup_function

from .event_handlers import handle_voice_client_join, handle_voice_client_move, handle_voice_client_update, \
    handle_voice_client_leave, handle_voice_server_update, handle_voice_client_ghost, handle_voice_client_shutdown

def setup_ext_solarlink(client):
    """
    Setups the solarlink extension on the client.
    
    Parameters
    ----------
    client : ``Client``
        The client to setup the extension into.
    
    Returns
    -------
    solarlink : ``SolarClient``
        The solar client controlling lavalink nodes.
    
    Raises
    ------
    TypeError
        - If `client` is not ``Client``.
    RuntimeError
        - If any attribute of the extension, is already used by the client.
    
    """
    for attr_name in ('solarlink', ):
        if hasattr(client, attr_name):
            raise RuntimeError(f'The client already has an attribute named as `{attr_name}`.')
    
    solar_client = SolarClient(client)
    client.solarlink = solar_client
    
    client.events(handle_voice_client_join, name = 'voice_client_join', overwrite=True)
    client.events(handle_voice_client_move, name = 'voice_client_move', overwrite=True)
    client.events(handle_voice_client_update, name = 'voice_client_update', overwrite=True)
    client.events(handle_voice_client_leave, name = 'voice_client_leave', overwrite=True)
    client.events(handle_voice_server_update, name = 'voice_server_update', overwrite=True)
    client.events(handle_voice_client_ghost, name = 'voice_client_ghost', overwrite=True)
    client.events(handle_voice_client_shutdown, name = 'voice_client_shutdown', overwrite=True)
    
    return solar_client


register_library_extension('HuyaneMatsu.solarlink')

register_setup_function(
    'HuyaneMatsu.solarlink',
    setup_ext_solarlink,
    None,
    None,
)



__all__ = ()

from ...emoji import Emoji, create_emoji_from_exclusive_data, put_exclusive_emoji_data_into
from ...field_parsers import force_string_parser_factory, bool_parser_factory
from ...field_putters import force_string_putter_factory, bool_optional_putter_factory
from ...field_validators import force_string_validator_factory, bool_validator_factory

from .constants import NAME_LENGTH_MIN, NAME_LENGTH_MAX

# emoji

def put_emoji_into(emoji, data, defaults):
    """
    Puts the `emoji`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The forum tag's emoji.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if defaults or (emoji is not None):
        put_exclusive_emoji_data_into(emoji, data)
    
    return data


def parse_emoji(data):
    """
    Parses out the `emoji` field from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Channel data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    return create_emoji_from_exclusive_data(data)


def validate_emoji(emoji):
    """
    Validates the given `emoji` field.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The emoji of the forum tag.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    
    Raises
    ------
    TypeError
        - If `emoji` is not `None`, ``Emoji``
    """
    if (emoji is not None) and (not isinstance(emoji, Emoji)):
        raise TypeError(
            f'`emoji` can be `None`, `{Emoji.__name__}`, got {emoji.__class__.__name__}; {emoji!r}.'
        )
    
    return emoji

# moderated

parse_moderated = bool_parser_factory('moderated', False)
put_moderated_into = bool_optional_putter_factory('moderated', False)
validate_moderated = bool_validator_factory('moderated')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



__all__ = ('AutoModerationRuleTriggerMetadataBase', )

from scarletio import RichAttributeErrorBaseType


class AutoModerationRuleTriggerMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for ``AutoModerationRule``'s trigger metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new trigger metadata instance.
        """
        return object.__new__(cls)
        
    def __repr__(self):
        """Returns the trigger metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new auto moderation rule trigger metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Auto moderation rule trigger metadata payload.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self):
        """
        Converts the trigger metadata to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two trigger metadatas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the trigger metadata's hash value."""
        return 0
    
    
    def copy(self):
        """
        Copies the trigger metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))




__all__ = ('ACTIVITY_UNKNOWN', 'Activity')

import warnings

from scarletio import RichAttributeErrorBaseType, copy_docs

from ..color import Color
from ..http import urls as module_urls
from ..preconverters import preconvert_preinstanced_type
from ..utils import DISCORD_EPOCH_START

from .constants import (
    ACTIVITY_COLOR_GAME, ACTIVITY_COLOR_NONE, ACTIVITY_COLOR_SPOTIFY, ACTIVITY_COLOR_STREAM, ACTIVITY_CUSTOM_IDS,
    ACTIVITY_CUSTOM_ID_DEFAULT
)
from .preinstanced import ActivityType
from .metadata import ActivityMetadataBase


class Activity(RichAttributeErrorBaseType):
    """
    Represents a Discord activity.
    
    Attributes
    ----------
    metadata : ``ActivityMetadataBase``
        Metadata of the activity containing extra fields about itself.
    type : ``ActivityType``
        The activity's type.
    """
    __slots__ = ('metadata', 'type')
    
    def __new__(cls, name, *, type_=None, **keyword_parameters):
        """
        Creates a new activity with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the activity.
        type_ : `None`, `int` = `None`, Optional (Keyword only)
            The type value of the activity.
        **keyword_parameters : Keyword parameters
            Additional parameters to pass to the activity-type specific constructor.
        
        Other Parameters
        ----------------
        application_id : `int`, Optional (Keyword only)
            The id of the activity's application.
        assets : `None`, ``ActivityAssets``, Optional (Keyword only)
             The activity's assets.
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the activity was created.
        details : `None`, `str`, Optional (Keyword only)
            What the player is currently doing.
        flags : ``ActivityFlag``, `int`, Optional (Keyword only)
            The flags of the activity.
        id_ : `int`, Optional (Keyword only)
            The id of the activity.
        party : `None`, ``ActivityParty``, Optional (Keyword only)
            The activity's party.
        secrets : `None`, ``ActivitySecret``, Optional (Keyword only)
            The activity's secrets.
        session_id : `None`, `str`, Optional (Keyword only)
            Spotify activity's session's id.
        state : `None`, `str`, Optional (Keyword only)
            The player's current party status.
        sync_id : `None`, `str`, Optional (Keyword only)
            The ID of the currently playing track of a spotify activity.
        timestamps : ``ActivityTimestamps``, Optional (Keyword only)
            The activity's timestamps.
        url : `None`, `str`, Optional (Keyword only)
            The url of the activity. Only twitch and youtube urls are supported.
        
        Returns
        -------
        activity : ``Activity``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - If extra or unused parameters were given.
        ValueError
            - If a parameter's value is incorrect.
        """
        if type_ is None:
            type_ = ActivityType.game
        else:
            type_ = preconvert_preinstanced_type(type_, 'type_', ActivityType)
        
        keyword_parameters['name'] = name
        metadata = type_.metadata_type(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        self = object.__new__(cls)
        self.metadata = metadata
        self.type = type_
        return self
    
    
    def __hash__(self):
        """Returns the activity's hash value."""
        hash_value = 0
        
        # metadata
        hash_value ^= hash(self.metadata)
        
        # type
        hash_value ^= self.type.value
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two activities are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # metadata
        if self.metadata != other.metadata:
            return False
        
        # type
        if self.type != other.type:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the activity's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append('~')
        repr_parts.append(repr(type_.value))
        
        # metadata
        repr_parts.append(', metadata=')
        repr_parts.append(repr(self.metadata))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an activity from the json data sent by Discord.
        
        If the data is `None` returns `ACTIVITY_UNKNOWN`.
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `Any`) items
            Activity data received from Discord.
        
        Returns
        -------
        activity : ``Activity``
        """
        if data is None:
            return ACTIVITY_UNKNOWN
        
        type_ = ActivityType.get(data.get('type', 0))
        metadata = type_.metadata_type.from_data(data)
        
        self = object.__new__(cls)
        self.metadata = metadata
        self.type = type_
        return self
    
    
    def to_data(self, *, include_internals=False, user=False):
        """
        Converts the activity to json serializable dictionary, which can be sent with bot account to change activity.
        
        Parameters
        ----------
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields, like id-s should be present as well.
        user : `bool` = `False`, Optional (Keyword only)
            Whether not only bot compatible fields should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = self.metadata.to_data(include_internals = include_internals, user = user)
        data['type'] = self.type.value
        
        if include_internals:
            # id | receive only?
            if ('id' not in data):
                data['id'] = self.discord_side_id
        
        return data
    
    
    def bot_dict(self):
        """
        Deprecated and will be removed in 2023 Jan, please use ``.to_data`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.bot_dict` is deprecated and will be removed in 2023 Jan. '
                f'Please use `.to_data` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.to_data()
    
    
    def user_dict(self):
        """
        Deprecated and will be removed in 2023 Jan, please use ``.to_data(user = True)`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.user_dict` is deprecated and will be removed in 2023 Jan. '
                f'Please use `.to_data(user = True)` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.to_data(user = True)
    
    
    def full_dict(self):
        """
        Deprecated and will be removed in 2023 Jan, please use ``.to_data(include_internals = True)`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.full_dict` is deprecated and will be removed in 2023 Jan. '
                f'Please use `.to_data(include_internals = True)` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.to_data(include_internals = True)
    
    
    def _update_attributes(self, data):
        """
        Updates the activity by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        type_ = ActivityType.get(data.get('type', 0))
        metadata_type = type_.metadata_type
        
        metadata = self.metadata
        
        if metadata_type is type(metadata):
            metadata._update_attributes(data)
        else:
            self.metadata = metadata_type.from_data(data)
        
        self.type = type_
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the activity and returns the changes in a `dict` of (`attribute-name`, `old-value`) items.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +-------------------+-----------------------------------+
        | Keys              | Values                            |
        +===================+===================================+
        | assets            | `None`, ``ActivityAssets``        |
        +-------------------+-----------------------------------+
        | created_at        | `None`, `datetime`                |
        +-------------------+-----------------------------------+
        | details           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | emoji             | `None`, ``Emoji``                 |
        +-------------------+-----------------------------------+
        | flags             | ``ActivityFlag``                  |
        +-------------------+-----------------------------------+
        | name              | `str`                             |
        +-------------------+-----------------------------------+
        | metadata          | ``ActivityMetadataBase``          |
        +-------------------+-----------------------------------+
        | party             | `None`, ``ActivityParty``         |
        +-------------------+-----------------------------------+
        | secrets           | `None`, ``ActivitySecrets``       |
        +-------------------+-----------------------------------+
        | session_id        | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | state             | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | sync_id           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | timestamps        | `None`, `ActivityTimestamps``     |
        +-------------------+-----------------------------------+
        | type              | ``ActivityType``                  |
        +-------------------+-----------------------------------+
        | url               | `None`, `str`                     |
        +-------------------+-----------------------------------+
        """
        type_ = ActivityType.get(data.get('type', 0))
        metadata_type = type_.metadata_type
        
        metadata = self.metadata
        
        if metadata_type is type(metadata):
            old_attributes = metadata._difference_update_attributes(data)
        
        else:
            old_attributes = {
                'metadata': metadata
            }
            
            self.metadata = metadata_type.from_data(data)
        
        if (type_ is not self.type):
            old_attributes['type'] = self.type
            self.type = type_
        
        return old_attributes
    
    
    # Field proxies
    
    @property
    @copy_docs(ActivityMetadataBase.application_id)
    def application_id(self):
        return self.metadata.application_id
    
    
    @property
    @copy_docs(ActivityMetadataBase.assets)
    def assets(self):
        return self.metadata.assets
        
    
    @property
    @copy_docs(ActivityMetadataBase.created_at)
    def created_at(self):
        created_at = self.metadata.created_at
        if (created_at is None):
            created_at = DISCORD_EPOCH_START
        
        return created_at
    
    
    @property
    @copy_docs(ActivityMetadataBase.details)
    def details(self):
        return self.metadata.details
    
    
    @property
    @copy_docs(ActivityMetadataBase.emoji)
    def emoji(self):
        return self.metadata.emoji
    
    
    @property
    @copy_docs(ActivityMetadataBase.flags)
    def flags(self):
        return self.metadata.flags
    
    
    @property
    @copy_docs(ActivityMetadataBase.id)
    def id(self):
        return self.metadata.id
    
    
    @property
    @copy_docs(ActivityMetadataBase.name)
    def name(self):
        return self.metadata.name
    
    
    @property
    @copy_docs(ActivityMetadataBase.party)
    def party(self):
        return self.metadata.party
    
    
    @property
    @copy_docs(ActivityMetadataBase.secrets)
    def secrets(self):
        return self.metadata.secrets
    
    
    @property
    @copy_docs(ActivityMetadataBase.session_id)
    def session_id(self):
        return self.metadata.session_id
    
    
    @property
    @copy_docs(ActivityMetadataBase.state)
    def state(self):
        return self.metadata.state
    
    
    @property
    @copy_docs(ActivityMetadataBase.sync_id)
    def sync_id(self):
        return self.metadata.sync_id
    
    
    @property
    @copy_docs(ActivityMetadataBase.timestamps)
    def timestamps(self):
        return self.metadata.timestamps
    
    
    @property
    @copy_docs(ActivityMetadataBase.url)
    def url(self):
        return self.metadata.url
    
    # utility
    
    @property
    def color(self):
        """
        Returns the activity's color.
        
        Returns
        -------
        color : ``Color``
        """
        type_ = self.type
        if (type_ is ActivityType.game):
            color = ACTIVITY_COLOR_GAME
            
        elif (type_ is ActivityType.custom):
            color = ACTIVITY_COLOR_NONE
        
        elif (type_ is ActivityType.stream):
            if (self.url is None):
                color = ACTIVITY_COLOR_GAME
            else:
                color = ACTIVITY_COLOR_STREAM
        
        elif (type_ is ActivityType.spotify):
            color = ACTIVITY_COLOR_SPOTIFY
        
        elif  (type_ is ActivityType.unknown):
            color = ACTIVITY_COLOR_NONE
        
        else:
            # Place holder for new activity types.
            # Right now covers: watching & competing
            color = ACTIVITY_COLOR_GAME
        
        return color
    
    
    @property
    def discord_side_id(self):
        """
        Returns the activity's Discord side id. If the activity implements id returns that, else tries to look it put
        from constants.
        
        Returns
        -------
        discord_side_id : `str`
        """
        id_ = self.id
        if id_:
            discord_side_id = format(self.id, 'x')
        else:
            discord_side_id = ACTIVITY_CUSTOM_IDS.get(self.type, ACTIVITY_CUSTOM_ID_DEFAULT)
        
        return discord_side_id
    

    
    @property
    def twitch_name(self):
        """
        If the user streams on twitch, returns it's twitch name.
        
        Only applicable for stream activities.
        
        Returns
        -------
        name : `None`, `str`
        """
        if self.type is not ActivityType.stream:
            return None
        
        assets = self.assets
        if assets is None:
            return None
        
        image_large = assets.image_large
        if image_large is None:
            return None
        
        if not image_large.startswith('twitch:'):
            return None
        
        return image_large[len('twitch:'):]
    
    
    @property
    def twitch_preview_image_url(self):
        """
        Returns the activity's twitch preview image url.
        
        Only applicable for stream activities.
        
        Returns
        -------
        preview_image_url : `None`, `str`
        """
        twitch_name = self.twitch_name
        if (twitch_name is not None):
            return f'https://static-cdn.jtvnw.net/previews-ttv/live_user_{twitch_name}.png'
    
    
    @property
    def youtube_video_id(self):
        """
        If the user streams on youtube, returns it's stream's respective video identifier.
        
        Only applicable for stream activities.
        
        Returns
        -------
        video_id : `None`, `str`
        """
        if self.type is not ActivityType.stream:
            return None
        
        assets = self.assets
        if assets is None:
            return None
        
        image_large = assets.image_large
        if image_large is None:
            return None
        
        if not image_large.startswith('youtube:'):
            return None
        
        return image_large[len('youtube:'):]
    
    
    @property
    def youtube_preview_image_url(self):
        """
        Returns the activity's youtube preview image url.
        
        Only applicable for stream activities.
        
        Returns
        -------
        preview_image_url : `None`, `str`
        """
        youtube_video_id = self.youtube_video_id
        if (youtube_video_id is not None):
            return f'https://i.ytimg.com/vi/{youtube_video_id}/hqdefault_live.jpg'
    
    
    @property
    def duration(self):
        """
        Drops a deprecation warning and returns ``.spotify_track_duration``.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.duration` is deprecated and will be removed in 2022 December. '
                f'Please use `.spotify_track_duration` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.spotify_track_duration
    
    
    @property
    def spotify_track_duration(self):
        """
        Returns the spotify activity's duration, or `None` if not applicable.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        duration : `None`, `timedelta`
        """
        if self.type is not ActivityType.spotify:
            return None
        
        timestamps = self.timestamps
        start = timestamps.start
        if start is None:
            return None
        
        end = timestamps.end
        if end is None:
            return None
        
        return end - start
    
    
    @property
    def spotify_cover_id(self):
        """
        If the user listens to spotify, returns it's spotify name.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        name : `None`, `str`
        """
        if self.type is not ActivityType.spotify:
            return None
        
        assets = self.assets
        if assets is None:
            return None
        
        image_large = assets.image_large
        if image_large is None:
            return None
        
        if not image_large.startswith('spotify:'):
            return None
        
        return image_large[len('spotify:'):]
    
    
    @property
    def spotify_album_cover_url(self):
        """
        Returns the spotify activity's currently playing track's album's cover url if applicable.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        album_cover_url : `None`, `str`
        """
        spotify_cover_id = self.spotify_cover_id
        if (spotify_cover_id is not None):
            return f'https://i.scdn.co/image/{spotify_cover_id}'
    
    
    @property
    def album_cover_url(self):
        """
        Drops a deprecation warning and returns ``.spotify_album_cover_url``.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.album_cover_url` is deprecated and will be removed in 2022 December. '
                f'Please use `.spotify_album_cover_url` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.spotify_album_cover_url
    
    
    @property
    def track_id(self):
        """
        Drops a deprecation warning and returns ``.spotify_album_cover_url``.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.track_id` is deprecated and will be removed in 2023 January. '
                f'Please use `.spotify_track_id` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.spotify_track_id
    
    
    @property
    def spotify_track_id(self):
        """
        Returns the song's identifier.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        track_id : `None`, `str`
        """
        if self.type is not ActivityType.spotify:
            return None
        
        return self.sync_id
    
    
    @property
    def track_url(self):
        """
        Drops a deprecation warning and returns ``.spotify_track_url``.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.track_url` is deprecated and will be removed in 2023 January. '
                f'Please use `.spotify_track_url` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.spotify_track_url
    
    
    @property
    def spotify_track_url(self):
        """
        Returns url to the spotify activity's song.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        url : `None`, `str`
        """
        spotify_track_id = self.spotify_track_id
        if (spotify_track_id is not None):
            return f'https://open.spotify.com/track/{spotify_track_id}'
    
    
    image_large_url = property(module_urls.activity_asset_image_large_url)
    image_large_url_as = module_urls.activity_asset_image_large_url_as
    image_small_url = property(module_urls.activity_asset_image_small_url)
    image_small_url_as = module_urls.activity_asset_image_small_url_as
    
    
    @property
    def start(self):
        """
        Returns when the activity was started if applicable.
        
        Returns
        -------
        start : `None`, `datetime`
        """
        timestamps = self.timestamps
        if (timestamps is not None):
            return timestamps.start
    
    
    @property
    def end(self):
        """
        Returns when the activity ended or will end if applicable.
        
        Returns
        -------
        start : `None`, `datetime`
        """
        timestamps = self.timestamps
        if (timestamps is not None):
            return timestamps.end


ACTIVITY_UNKNOWN = Activity('', type_=ActivityType.unknown)



__all__ = ('AnsiBackgroundColor', 'AnsiForegroundColor', 'AnsiTextDecoration', 'create_ansi_format_code')

from scarletio import copy_docs
from .bases import Preinstance as P, PreinstancedBase
from .color import Color


class AnsiTextDecoration(PreinstancedBase):
    """
    Represents an ansi text decoration.
    
    Attributes
    ----------
    value : `int`
        The unique identifier of the text decoration.
    name : `str`
        The default name of the text decoration.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AnsiTextDecoration``) items
        The predefined text decorations stored in `.value` - `object` relation.
    VALUE_TYPE : `type` = `int`
        Text decoration code's type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name to use as the preinstanced objects'.
    
    Each text decoration is stored as a class attribute:
    
    +-----------------------+-----------+-----------------------+
    | Class Attribute name  | value     | name                  |
    +=======================+===========+=======================+
    | none                  | 0         | none                  |
    +-----------------------+-----------+-----------------------+
    | bold                  | 1         | bold                  |
    +-----------------------+-----------+-----------------------+
    | underline             | 4         | underline             |
    +-----------------------+-----------+-----------------------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    none = P(0, 'none')
    bold = P(1, 'bold')
    underline = P(4, 'underline')


class AnsiFormatColor(PreinstancedBase):
    """
    Represents an ansi color format code.
    
    Attributes
    ----------
    value : `int`
        The unique identifier of the format color.
    name : `str`
        The default name of the format color.
    color : ``Color``
        The real color value.
    color_name : `str`
        A more accurate name of the color.
    
    Class Attributes
    ----------------
    INSTANCES : `NoneType` = `NotImplemented`
        The instances of the preinstanced type. Subclasses should overwrite it as `dict`.
    VALUE_TYPE : `type` = `int`
        Format color code's type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name to use as the preinstanced objects'.
    """
    __slots__ = ('color', 'color_name')
    
    VALUE_TYPE = int
    
    @classmethod
    @copy_docs(PreinstancedBase._from_value)
    def _from_value(cls, value):
        raise NotImplementedError
    
    
    def __init__(self, value, name, color, color_name):
        """
        Creates a new format color with the given parameters and stores it at the class's `.INSTANCES`.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the format color.
        name : `str`
            The format color's name.
        color : ``Color``
            The real color value.
        """
        self.name = name
        self.value = value
        self.color = color
        self.color_name = color_name
        self.INSTANCES[value] = self



class AnsiBackgroundColor(AnsiFormatColor):
    """
    Represents an ansi background color code.
    
    Attributes
    ----------
    value : `int`
        The unique identifier of the format color.
    name : `str`
        The default name of the format color.
    color : ``Color``
        The real color value.
    color_name : `str`
        A more accurate name of the color.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AnsiBackgroundColor``) items
        The predefined background colors stored in `.value` - `object` relation.
    VALUE_TYPE : `type` = `int`
        Format color code's type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name to use as the preinstanced objects'.
    
    Each background color is stored as a class attribute:
    
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | Class Attribute name  | value     | name                  | color     | color name        | Additional notes              |
    +=======================+===========+=======================+===========+===================+===============================+
    | black                 | 40        | black                 | 002b36    | daintree          | Dark cyan, almost black.      |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | red                   | 41        | red                   | 0xcb4b16  | orange roughy     | Bright ~ brick red.           |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | marble_blue_cyan      | 42        | marble blue cyan      | 0x586e75  | cutty sark        | Looks gray to me.             |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | marble_blue           | 43        | marble blue           | 0x657b83  | pale sky          | Looks gray to me.             |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | gray                  | 44        | gray                  | 0x657b83  | oslo gray         | Looks gray to me.             |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | blue                  | 45        | blue                  | 0x6c71c4  | blue marguerite   | Blue-violet                   |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | silver                | 46        | silver                | 0x93a1a1  | pewter            | Looks gray to me.             |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | white                 | 47        | old lace              | 0xfdf6e3  | old lace          | Pale yellow-orange.           |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    
    black = P(40, 'black', Color(0x002b36), 'daintree')
    red = P(41, 'red', Color(0xcb4b16), 'orange roughy')
    marble_blue_cyan = P(42, 'marble blue cyan', Color(0x586e75), 'cutty sark')
    marble_blue = P(43, 'marble blue', Color(0x657b83), 'pale sky')
    gray = P(44, 'gray', Color(0x839496), 'oslo gray')
    blue = P(45, 'blue', Color(0x6c71c4), 'blue marguerite')
    silver = P(46, 'silver', Color(0x93a1a1), 'pewter')
    white = P(47, 'white', Color(0xfdf6e3), 'old lace')


class AnsiForegroundColor(AnsiFormatColor):
    """
    Represents an ansi foreground color code.
    
    Attributes
    ----------
    value : `int`
        The unique identifier of the format color.
    name : `str`
        The default name of the format color.
    color : ``Color``
        The real color value.
    color_name : `str`
        A more accurate name of the color.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AnsiForegroundColor``) items
        The predefined foreground colors stored in `.value` - `object` relation.
    VALUE_TYPE : `type` = `int`
        Format color code's type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name to use as the preinstanced objects'.
        
    Each foreground color is stored as a class attribute:
    
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | Class Attribute name  | value     | name                  | color     | color name        | Additional notes              |
    +=======================+===========+=======================+===========+===================+===============================+
    | black                 | 30        | black                 | 0x073642  | tiber             | Dark cyan, almost black.      |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | red                   | 31        | red                   | 0xdc322f  | punch             | Red.                          |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | green                 | 32        | green                 | 0x859900  | limeade           | Darker lime shade.            |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | orange                | 33        | orange                | 0xb58900  | pirate gold       | Warm orange.                  |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | blue                  | 34        | blue                  | 0x268bd2  | curious blue      | Blue.                         |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | pink                  | 35        | pink                  | 0xd33682  | cerise            | Red-pink.                     |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | teal                  | 36        | teal                  | 0x2aa198  | jungle green      | Teal (greenish blue).         |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    | white                 | 37        | white                 | 0xffffff  | white             | Black on bright theme.        |
    +-----------------------+-----------+-----------------------+-----------+-------------------+-------------------------------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    
    black = P(30, 'black', Color(0x073642), 'tiber')
    red = P(31, 'red', Color(0xdc322f), 'punch')
    green = P(32, 'green', Color(0x859900), 'limeade')
    orange = P(33, 'orange', Color(0xb58900), 'pirate gold')
    blue = P(34, 'blue', Color(0x268bd2), 'curious blue')
    pink = P(35, 'pink', Color(0xd33682), 'cerise')
    teal = P(36, 'teal', Color(0x2aa198), 'jungle green')
    white = P(37, 'white', Color(0xffffff), 'white')


def create_ansi_format_code(text_decoration = None, background_color = None, foreground_color = None):
    """
    Creates an ansi text format code for `ansi` codeblocks.
    
    > If no parameter is given, will generate a format reset code.
    
    Parameters
    ----------
    text_decoration : `None`, ``AnsiTextDecoration`` = `None`, Optional
        Text decoration.
    background_color : `None`, ``AnsiBackgroundColor`` = `None`, Optional
        background color.
    foreground_color : `None`, ``AnsiForegroundColor`` = `None`, Optional
        Foreground color.
    
    Returns
    -------
    format_code : `str`
    """
    format_code_parst = ['\u001b[']
    
    field_added = False
    
    if (text_decoration is not None):
        field_added = True
        
        format_code_parst.append(str(text_decoration.value))
    
    if (background_color is not None):
        if field_added:
            format_code_parst.append(';')
        else:
            field_added = True
        
        format_code_parst.append(str(background_color.value))
    
    if (foreground_color is not None):
        if field_added:
            format_code_parst.append(';')
        else:
            field_added = True
        
        format_code_parst.append(str(foreground_color.value))
    
    if not field_added:
        format_code_parst.append('0')
    
    format_code_parst.append('m')
    
    return ''.join(format_code_parst)



__all__ = ('IntegrationApplication',)

from ...bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ...http import urls as module_urls
from ...preconverters import preconvert_snowflake
from ...user import ZEROUSER

from .fields import (
    parse_bot, parse_description, parse_name, put_bot_into, put_description_into, put_name_into, validate_bot,
    validate_description, validate_name
)


class IntegrationApplication(DiscordEntity):
    """
    Represents a Discord ``Application`` received with Integration data.
    
    Attributes
    ----------
    id : `int`
        The application's id.
    icon_hash : `int`
        The application's icon's hash as `uint128`.
    icon_type : `IconType`
        The application's icon's type.
    bot : ``ClientUserBase``
        The application's bot if applicable.
    description : `None` `str`
        The description of the application. Defaults to empty string.
    name : `str`
        The name of the application. Defaults to empty string.
    """
    __slots__ = ('bot', 'description', 'name', )
    
    icon = IconSlot('icon', 'icon', module_urls.application_icon_url, module_urls.application_icon_url_as)
    
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a new integration application from the given keyword parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        bot : `None`, ``ClientUserBase``, Optional (Keyword only)
            The application's bot if applicable.
        
        icon : `None`, `bytes`, `bytearray`, `memoryview`, ``Icon``, Optional (Keyword only)
            The icon of the integration application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        name : `None, `str`, Optional (Keyword only)
            The name of the application.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        self = cls._create_empty(0)
        self._set_attributes_from_keyword_parameters(keyword_parameters)
        
        # icon
        self.icon = cls.icon.parse_from_keyword_parameters(keyword_parameters, allow_data = True)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        return self
    
    
    @classmethod
    def precreate(cls, integration_application_id, **keyword_parameters):
        """
        Creates a new integration application with the given predefined fields.
        
        > Since integration applications are not globally cached, this method is only used for testing.
        
        Parameters
        ----------
        integration_application_id : `int`
            The integration application's id.
        **keyword_parameters : Keyword parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        bot : `None`, ``ClientUserBase``, Optional (Keyword only)
            The application's bot if applicable.
        
        icon : `None`, ``Icon``, Optional (Keyword only)
            The icon of the integration application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        name : `None, `str`, Optional (Keyword only)
            The name of the application.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        integration_application_id = preconvert_snowflake(integration_application_id, 'integration_application_id')
        self = cls._create_empty(integration_application_id)
        self._set_attributes_from_keyword_parameters(keyword_parameters)
        
        # icon
        self.icon = cls.icon.parse_from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        return self
    
    
    @classmethod
    def _create_empty(cls, integration_application_id):
        """
        Creates a new integration application with it's defaults attributes set.
        
        Parameters
        ----------
        integration_application_id : `int`
            The integration application's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = integration_application_id
        self.bot = ZEROUSER
        self.description = None
        self.icon_type = ICON_TYPE_NONE
        self.icon_hash = 0
        self.name = ''
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new integration application instance with the given application data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Application data included within integration payload.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = int(data['id'])
        self._set_icon(data)
        self.bot = parse_bot(data)
        self.description = parse_description(data)
        self.name = parse_name(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the integration application to json serializable data.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # id
        if include_internals:
            data['id'] = str(self.id)
        
        # bot
        put_bot_into(self.bot, data, defaults)
        
        # description
        put_description_into(self.description, data, defaults)
        
        # icon
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        
        # name
        put_name_into(self.name, data, defaults)
        
        
        return data
    
    
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        """
        Sets the integration application's attributes from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            The keyword parameters to set the integration application's attributes from.
        
        Raises
        ------
        TypeError
            A field of invalid type.
        ValueError
            A field of invalid value.
        """
        # bot
        try:
            bot = keyword_parameters.pop('bot')
        except KeyError:
            pass
        else:
            self.bot = validate_bot(bot)
        
        # description
        try:
            description = keyword_parameters.pop('description')
        except KeyError:
            pass
        else:
            self.description = validate_description(description)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            self.name = validate_name(name)
    
    
    @property
    def partial(self):
        """
        Returns whether the integration application is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return (self.id == 0)
    
    
    def __eq__(self, other):
        """Returns whether the two integration applications are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two integration applications not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        if self.id != other.id:
            return False
        
        # bot
        if self.bot != other.bot:
            return False
        
        # icon_hash
        if self.icon_hash != other.icon_hash:
            return False
        
        # icon_type
        if self.icon_type != other.icon_type:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the integration application's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        integration_application_id = self.id
        if integration_application_id:
            repr_parts.append(' id=')
            repr_parts.append(repr(integration_application_id))
        else:
            repr_parts.append(' (partial)')
        
        name = self.name
        if name:
            repr_parts.append(', name=')
            repr_parts.append(repr(name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the integration application's hash value."""
        # These entities are not cached, so we wont use their `id` if applicable.
        hash_value = 0
        
        # bot
        hash_value ^= hash(self.bot)
        
        # icon
        hash_value ^= hash(self.icon)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # name
        name = self.name
        if (description is None) or (description != name):
            hash_value ^= hash(name)
        
        return hash_value



__all__ = ('calculate_snapshot_difference', 'revert_snapshot', 'take_snapshot',)

from ....discord.core import CLIENTS

from .base_snapshot_type import SNAPSHOT_TYPES


def take_snapshot():
    """
    Takes a snapshot of the clients.
    
    Returns
    -------
    snapshots : `list` of ``BaseSnapshotType``
    """
    snapshots = []
    
    for client in CLIENTS.values():
        for snapshot_type in SNAPSHOT_TYPES:
            snapshot = snapshot_type(client)
            snapshots.append(snapshot)
    
    return snapshots


def _build_snapshot_tree(snapshots):
    """
    Builds a tree fro the given snapshot.
    
    Parameters
    ----------
    snapshots : `list` of ``BaseSnapshotType``
        The snapshot to build tree from.
    
    Returns
    -------
    snapshot_tree : `dict` of (``client``, `dict` of (`type`, ``BaseSnapshotType``) items) items
    """
    snapshot_tree = {}
    
    for snapshot in snapshots:
        client = snapshot.client
        if client is None:
            continue
        
        try:
            by_client = snapshot_tree[client]
        except KeyError:
            by_client = {}
            snapshot_tree[client] = by_client
        
        by_client[type(snapshot)] = snapshot
    
    return snapshot_tree


def calculate_snapshot_difference(snapshots_1, snapshots_2):
    """
    Calculates snapshot differences between two snapshots.
    
    Parameters
    ----------
    snapshots_1 : `list` of ``BaseSnapshotType``
        Old taken snapshot.
    snapshots_2 : `list` of ``BaseSnapshotType``
        New taken snapshot.
    
    Returns
    -------
    new_snapshots : `list` of ``BaseSnapshotType``
    """
    tree_1 = _build_snapshot_tree(snapshots_1)
    tree_2 = _build_snapshot_tree(snapshots_2)
    
    new_snapshots = []
    
    for client in {*tree_1.keys(), *tree_2.keys()}:
        try:
            snapshots_by_type_1 = tree_1[client]
        except KeyError:
            continue
        
        try:
            snapshots_by_type_2 = tree_2[client]
        except KeyError:
            continue
        
        for snapshot_type in {*snapshots_by_type_1, *snapshots_by_type_2}:
            try:
                snapshot_1 = snapshots_by_type_1[snapshot_type]
            except KeyError:
                continue
            
            try:
                snapshot_2 = snapshots_by_type_2[snapshot_type]
            except KeyError:
                continue
            
            new_snapshot = snapshot_1 - snapshot_2
            new_snapshots.append(new_snapshot)
    
    return new_snapshots


def revert_snapshot(snapshots):
    """
    Reverts the clients to a previous state based on snapshots.
    
    Parameters
    ----------
    snapshots : `list` of ``BaseSnapshotType``
        DSnapshots to revert
    """
    for snapshot in snapshots:
        if snapshot.is_revertible():
            snapshot.revert()



__all__ = ('add_default_plugin_variables',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.add_default_variables)
def add_default_plugin_variables(**variables):
    return PLUGIN_LOADER.add_default_variables(**variables)



__all__ = ()

import re, reprlib

from scarletio import export

from .builder_html import create_relative_link, description_serializer, graved_to_escaped, sub_section_serializer
from .graver import GravedAttributeDescription, GravedDescription
from .module_mapper import FunctionUnit, InstanceAttributeUnit, MAPPED_OBJECTS, ModuleUnit, PropertyUnit, TypeUnit
from .parser import ATTRIBUTE_NAME_RP, ATTRIBUTE_SECTION_NAME_RP

from html import escape as html_escape


class Structure:
    """
    Represents an extended docstring's html structure.
    
    Attributes
    ----------
    title : `str`
        The title of the represented unit or section.
    prefixed_title : `str`
        Can be used as local reference in the html file.
    children : `None` or (`list` of ``Structure``)
        The child units of sections. Set as `None` if would have non.
    """
    __slots__ = ( 'title', 'prefixed_title', 'children')
    
    def __init__(self, title, prefixed_title, children):
        """
        Creates a nw structure object with the given parameters.
        
        Parameters
        ----------
        title : `str`
            The title of the represented unit or section.
        prefixed_title : `str`
            Can be used as local reference in the html file.
        children : `None` or (`list` of ``Structure``)
            The child units of sections.
            
            Should be given as `None` if the structure has no child.
        """
        self.title = title
        self.prefixed_title = prefixed_title
        self.children = children


def create_relative_sectioned_link(source, target):
    """
    Creates relative link between two objects. Not like ``create_relative_link``, this function uses anchors as well.
    
    Parameters
    ----------
    source : ``QualPath``
        Source object's path.
    target : ``QualPath``
        Target object's path.
    
    Returns
    -------
    url : `str`
        Generated relative url.
    """
    while True:
        parent_maybe = source.parent
        unit = MAPPED_OBJECTS.get(parent_maybe, None)
        if unit is None:
            break
        
        if type(unit) is ModuleUnit:
            break
        
        source = parent_maybe
        continue
    
    target_backfetched = []
    
    while True:
        target_maybe = target.parent
        unit = MAPPED_OBJECTS.get(target_maybe, None)
        if unit is None:
            break
        
        if type(unit) is ModuleUnit:
            break
        
        target_backfetched.append(target.parts[-1])
        target = target_maybe
        continue
    
    url = create_relative_link(source, target)
    if target_backfetched:
        target_backfetched.reverse()
        target_backfetched = '-'.join(target_backfetched)
        target_backfetched = anchor_escape(target_backfetched)
        
        url = f'{url}#{target_backfetched}'
    
    return url

def anchor_escape(text):
    """
    Anchor escapes the given text.
    
    Parameters
    ----------
    text : `str`
        The text to escape.
    
    Returns
    -------
    escaped : `str`
    """
    return text.lower().replace(' ', '-')

ANCHOR_SVG = (
    '<svg viewBox="0 0 16 16">'
        '<path fill-rule="evenodd" d="M 7.775 3.275 a 0.75 0.75 0 0 0 1.06 1.06 l 1.25 -1.25 a 2 2 0 1 1 2.83 2.83 '
            'l -2.5 2.5 a 2 2 0 0 1 -2.83 0 a 0.75 0.75 0 0 0 -1.06 1.06 a 3.5 3.5 0 0 0 4.95 0 l 2.5 -2.5 '
            'a 3.5 3.5 0 0 0 -4.95 -4.95 l -1.25 1.25 Z m -4.69 9.64 a 2 2 0 0 1 0 -2.83 l 2.5 -2.5 a 2 2 0 0 1 2.83 0 '
            'a 0.75 0.75 0 0 0 1.06 -1.06 a 3.5 3.5 0 0 0 -4.95 0 l -2.5 2.5 a 3.5 3.5 0 0 0 4.95 4.95 l 1.25 -1.25 '
            'a 0.75 0.75 0 0 0 -1.06 -1.06 l -1.25 1.25 a 2 2 0 0 1 -2.83 0 Z"></path>'
    '</svg>'
)

YES_SVG = (
    '<svg '
        'width="1.2em" '
        'height="1.2em" '
        'preserveAspectRatio="xMidYMid meet" '
        'viewBox="0 0 24 24" '
        'aria-hidden="true"'
    '>'
        '<g fill="none">'
            '<path '
                'd="M5 13l4 4L19 7" '
                'stroke="currentColor" '
                'stroke-width="2" '
                'stroke-linecap="round" '
                'stroke-linejoin="round"'
            '>'
            '</path>'
        '</g>'
    '</svg>'
)

def anchor_for_serializer(section_name):
    """
    Returns the anchor for the given section name.
    
    This function is a generator.
    
    Parameters
    ----------
    section_name : `str`
        The section's name.
    
    Yields
    ------
    html : `str`
    """
    yield '<a name="'
    yield section_name
    yield '" class="anchor" aria-hidden="true" href="#'
    yield section_name
    yield '">'
    yield ANCHOR_SVG
    yield '</a>'

def section_title_serializer(title):
    """
    Serializes to html the given section name.
    
    This function is a generator.
    
    Parameters
    ----------
    title : `str`
        The respective section's name.
    
    Yields
    -------
    html : `str`
    """
    yield '<h2>'
    yield from anchor_for_serializer(title)
    yield html_escape(title)
    yield '<div class="underline"></div></h2>'

SECTION_TYPE_DEFAULT = 0

SECTION_TYPE_USAGES = 10

SECTION_TYPE_INSTANCE_ATTRIBUTES = 20
SECTION_TYPE_X_ATTRIBUTES = 21
SECTION_TYPE_CLASS_ATTRIBUTES = 22

SECTION_TYPE_PARAMETERS = 30
SECTION_TYPE_OTHER_PARAMETERS = 31

SECTION_TYPE_YIELDS = 40
SECTION_TYPE_RETURNS = 41
SECTION_TYPE_RAISES = 42

SECTION_TYPE_NOTES = 50
SECTION_TYPE_X = 51

SECTION_TYPE_PROPERTIES = 60
SECTION_TYPE_METHODS = 61
SECTION_TYPE_FUNCTIONS = 62
SECTION_TYPE_CLASSES = 63
SECTION_TYPE_MODULES = 64

def get_attribute_section_mentioned_names(section_content):
    """
    Gets the attribute names mention by the respective section.
    
    Parameters
    ----------
    section_content : `list` of `Any`
    
    Returns
    -------
    names : `set` of `str`
        The parsed names from the respective section.
    """
    names = set()
    
    index = 0
    limit = len(section_content)
    while True:
        if index == limit:
            break
        
        part = section_content[index]
        if (type(part) is not GravedDescription):
            break
        
        starter = part.content[0]
        if (type(starter) is not str):
            break
        
        parsed = ATTRIBUTE_NAME_RP.match(starter)
        if parsed is None:
            break
        
        attr_name = parsed.group(1)
        
        index += 1
        if index != limit:
            part = section_content[index]
            if type(part) is list:
                index += 1
        
        names.add(attr_name)
        continue
    
    return names


def get_anchor_prefix_for(path):
    """
    Returns the given object's anchor prefix to remove overlapping anchor names.
    
    Parameters
    ----------
    path : ``QualPath``
        The respective unit.
    
    Returns
    -------
    prefix : `str`
    """
    prefix_parts = []
    while True:
        parent = path.parent
        unit = MAPPED_OBJECTS.get(parent, None)
        if unit is None:
            break
        
        if type(unit) is ModuleUnit:
            break
        
        prefix_parts.append(path.parts[-1])
        path = parent
        continue
    
    return anchor_escape('-'.join(prefix_parts))

def get_tier_for(path):
    """
    Returns the given object's tier level. Tier level should be used to decide header type.
    
    Parameters
    ----------
    path : ``QualPath``
        The respective unit's path.
    
    Returns
    -------
    tier : `int`
    """
    tier = 1
    while True:
        parent = path.parent
        unit = MAPPED_OBJECTS.get(parent, None)
        if unit is None:
            break
        
        if type(unit) is ModuleUnit:
            break
        
        tier += 2
        path = parent
        continue
    
    return tier

def get_parent_path_of(path):
    """
    Returns the given path 0 tier parent path.
    
    Parameters
    ----------
    path : ``QualPath``
        The respective unit's path
    
    Returns
    -------
    parent_path : ``QualPath``
    """
    while True:
        parent = path.parent
        unit = MAPPED_OBJECTS.get(parent, None)
        if unit is None:
            break
        
        if type(unit) is ModuleUnit:
            break
        
        path = parent
        continue
    
    return path

def name_sort_key(name):
    """
    Gets sort key for the given name, when sorting attribute, method, or such names.
    
    Parameters
    ----------
    name : `str`
            Name of an attribute, method or class.
    
    Returns
    -------
    key : `tuple` (`int`, `str`)
        The generated key.
    """
    if name.startswith('__'):
        if name == '__new__':
            prefix_count = 0
        elif name == '__init__':
             prefix_count = 1
        elif name == '__call__':
            prefix_count = 2
        else:
            prefix_count = 102
    elif name.startswith('_'):
        prefix_count = 101
    else:
        prefix_count = 100
    
    return prefix_count, name

def item_sort_key(item):
    """
    Gets sort key for the given item, when sorting name - docstring pairs
    
    Parameters
    ----------
    item : `tuple (`str`, ``DocString``)
        A `name` - ``Docstring`` relation.
    
    Returns
    -------
    key : `tuple` (`int`, `str`)
        The generated key.
    """
    return name_sort_key(item[0])


class SimpleSection:
    """
    Represents a simple section.
    
    Attributes
    ----------
    content : `list` of `Any`
        Contained section part.
    object : ``UnitBase``
        The owner unit.
    title : `None`, `str`
        The title of the section.
    path : ``QualPath``
        Path to use instead of the objects'.
    """
    __slots__ = ('content', 'object', 'title', 'path')
    
    def __init__(self, title, content, object_, path):
        """
        Creates a new attribute section with the given parameters.
        
        Parameters
        ----------
        title : `None`, `str`
            The title of the represented section.
        content : `list` of `Any`
            Contained section part.
        object_ : ``TypeUnit``
            The owner type-unit.
        path : `None`, ``QualPath``
            Path to use instead of the objects'.
        """
        self.title = title
        self.content = content
        self.object = object_
        if path is None:
            path = object_.path
        self.path = path
    
    def serialize(self):
        """
        Serializes the section to html string parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        object_ = self.object
        path = self.path
        title = self.title
        
        if (title is not None):
            prefix = get_anchor_prefix_for(path)
            if prefix:
                prefix = f'{prefix}-{anchor_escape(title)}'
            else:
                prefix = anchor_escape(title)
            
            tier = get_tier_for(path) + 1
            
            yield '<h'
            yield str(tier)
            yield '>'
            yield from anchor_for_serializer(prefix)
            yield html_escape(title)
            yield '<div class="underline"></div></h'
            yield str(tier)
            yield '>'
        
        
        yield from sub_section_serializer(self.content, object_, get_parent_path_of(path),
            create_relative_sectioned_link)
        
        return
    
    def structurize(self):
        """
        Returns the structure of the section.
        
        Returns
        -------
        structure : `None`, ``Structure``
            If the section is unnamed, returns `None`.
        """
        title = self.title
        if title is None:
            return None
        
        prefixed_title = get_anchor_prefix_for(self.path)
        if prefixed_title:
            prefixed_title = f'{prefixed_title}-{anchor_escape(title)}'
        else:
            prefixed_title = anchor_escape(title)
        
        return Structure(title, prefixed_title, None)


class FunctionOrPropertySerializer:
    """
    Serializer for a method or for a property.
    
    Attributes
    ----------
    content : `list` of `Any`
        Contained section part.
    object : ``PropertyUnit``, ``FunctionUnit``
        The owner type-unit.
    parameter_section : `None`, ``ParameterSection``
        The parameter section of the serializer.
    path : ``QualPath``
        Path to use instead of the object's.
    """
    __slots__ = ('content', 'object', 'parameter_section', 'path')
    
    def __init__(self, object_, path=None):
        """
        Creates a new method or property serializer.
        
        Parameters
        ----------
        object_ : ``PropertyUnit``, ``FunctionUnit``
            The object to serialize.
        path : `None`, ``QualPath`` = `None`, Optional
            Path to use instead of the object's.
        """
        self.object = object_
        if path is None:
            path = object_.path
        self.path = path
        
        docs = object_.docs
        if docs is None:
            self.content = None
            self.parameter_section = None
            return
        
        parameter_section = None
        
        section_parts = {}
        for section_name, section_content in docs.sections:
            if section_name is None:
                section_parts[SECTION_TYPE_DEFAULT] = SimpleSection(None, section_content, object_, path)
                continue
            
            try:
                section_type_value = SECTION_NAME_TYPE_DEFAULT_RELATIONS[section_name]
            except KeyError:
                pass
            else:
                section_parts[section_type_value] = SimpleSection(section_name, section_content, object_, path)
                continue
            
            if section_name in ('Parameters', 'Other Parameters'):
                if (parameter_section is None):
                    parameter_section = ParameterSection(section_content, object_, path)
                    section_parts[SECTION_TYPE_PARAMETERS] = parameter_section
                else:
                    parameter_section.extend(section_content)
                
                continue
            
            try:
                existing = section_parts[SECTION_TYPE_X]
            except KeyError:
                existing = section_parts[SECTION_TYPE_X] = []
            
            existing.append(SimpleSection(section_name, section_content, object_, path))
            continue
        
        self.parameter_section = parameter_section
        
        content = []
        for part in sorted(section_parts.items()):
            sub_section = part[1]
            if type(sub_section) is list:
                content.extend(sub_section)
            else:
                content.append(sub_section)
        
        self.content = content
    
    
    def serialize(self):
        """
        Serializes the represented method or property to html parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        path = self.path
        
        name = path.parts[-1]
        prefix = get_anchor_prefix_for(path)
        tier = get_tier_for(path)
        
        yield '<h'
        yield str(tier)
        yield '>'
        yield from anchor_for_serializer(anchor_escape(prefix))
        yield html_escape(name)
        
        is_function = isinstance(self.object, FunctionUnit)
        
        if is_function:
            yield '<span class="function_parameter">('
        
            parameter_section = self.parameter_section
            if (parameter_section is not None):
                parameter_added = False
                
                for parameter_sub_section in parameter_section.parameter_sub_sections:
                    if parameter_sub_section.optional:
                        if parameter_added:
                            yield ', '
                        
                        yield '...'
                        break
                        
                    if parameter_added:
                        yield ', '
                    else:
                        parameter_added = True
                    
                    yield html_escape(parameter_sub_section.name)
            
            yield ')</span>'
        
        
        yield '</h'
        yield str(tier)
        yield '>'
        
        yield '<div class="underline"></div>'
        
        content = self.content
        if content is None:
            return
        
        if tier > 1:
            yield '<div class="sub_unit">'
        
        for serializer in content:
            yield from serializer.serialize()
        
        if tier > 1:
            yield '</div>'
        
        return
    
    def structurize(self):
        """
        Creates a structure for the represented unit.
        
        Returns
        -------
        structure : ``Structure``
        """
        content = self.content
        if content is None:
            children = None
        else:
            children = []
            for serializer in self.content:
                child = serializer.structurize()
                if child is None:
                    continue
                
                children.append(child)
                continue
            
            if not children:
                children = None
        
        path = self.path
        title = path.parts[-1]
        prefixed_title = get_anchor_prefix_for(path)
        return Structure(title, prefixed_title, children)


class UnitSection:
    """
    Represents a section, which is filled with units.
    
    Attributes
    ----------
    title : `str`
        The section's title.
    object : ``TypeUnit``
        The represented parent object.
    units : `list` of `tuple` (`str`, ``UnitBase``)
        The contained units.
    path : ``QualPath``
        Path to use instead of the object's.
    """
    __slots__ = ('title', 'object', 'units', 'path')
    
    def __init__(self, title, object_, units, path):
        """
        Creates a new unit section.
        
        Parameters
        ----------
        title : `str`
            The section's title.
        object_ : ``TypeUnit``
            The represented parent object.
        units : `list` of ``UnitBase``
            The contained units.
        path : `None`, ``QualPath``
            Path to use instead of the object's.
        """
        self.title = title
        self.object = object_
        self.units = units
        if path is None:
            path = object_.path
        self.path = path
    
    def serialize(self):
        """
        Serializes the unit section to html parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        title = self.title
        path = self.path
        prefix = get_anchor_prefix_for(path)
        
        if prefix:
            prefix = f'{prefix}-{anchor_escape(title)}'
        else:
            prefix = anchor_escape(title)
        
        tier = get_tier_for(path) + 1
        
        yield '<h'
        yield str(tier)
        yield '>'
        yield from anchor_for_serializer(prefix)
        yield html_escape(title)
        yield '<div class="underline"></div></h'
        yield str(tier)
        yield '>'
        
        yield '<div class="unit_section">'
        for name, unit in self.units:
            serializer_type = UNIT_CONVERSION_TABLE[type(unit)]
            serializer = serializer_type(unit, path / name)
            yield from serializer.serialize()
        
        yield '</div>'
        
        return
    
    def structurize(self):
        """
        Returns the structure of the unit section.
        
        Returns
        -------
        structure : `None`, ``Structure``
            If the section is unnamed, returns `None`.
        """
        path = self.path
        children = []
        
        for name, unit in self.units:
            serializer_type = UNIT_CONVERSION_TABLE[type(unit)]
            serializer = serializer_type(unit, path / name)
            child = serializer.structurize()
            children.append(child)
        
        if not children:
            children = None
        
        title = self.title
        prefixed_title = get_anchor_prefix_for(path)
        
        if prefixed_title:
            prefixed_title = f'{prefixed_title}-{anchor_escape(title)}'
        else:
            prefixed_title = anchor_escape(title)
        
        return Structure(title, prefixed_title, children)


class AttributeSection:
    """
    Represents an attribute section.
    
    Attributes
    ----------
    extra : `None`, `list` of `Any`
        Extra content after the attribute section.
    relations : `dict` of (`str`, `DocString`) items
        Attribute name, DocString relation.
    object : ``TypeUnit``
        The owner type-unit.
    title : `str`
        The title of the attribute section.
    path : ``QualPath``
        Path to use instead of the object's.
    """
    __slots__ = ('extra', 'relations', 'object', 'title', 'path')
    
    def __init__(self, title, mentioned_names, object_, path):
        """
        Creates a new attribute section with the given parameters.
        
        Parameters
        ----------
        title : `str`
            The title of the represented section.
        mentioned_names : `set` of `str`
            Mentioned attribute names at the attribute section.
        object_ : ``TypeUnit``
            The owner type-unit.
        path : `None`, ``QualPath``
            Path to use instead of the object's.
        """
        self.title = title
        docs = object_.docs
        if docs is None:
            relations = {mentioned_name:None for mentioned_name in mentioned_names}
            extra = None
        else:
            relations = {
                mentioned_name: docs.attribute_docstring_for(mentioned_name) for mentioned_name in mentioned_names
            }
            extra = docs.extra_attribute_docstring_for(title)
        self.relations = relations
        self.extra = extra
        self.object = object_
        if path is None:
            path = object_.path
        self.path = path
    
    def contains(self, name):
        """
        Returns whether the attribute section contains the given attribute name.
        """
        if name in self.relations:
            return True
        
        return False
    
    def add(self, name):
        """
        Adds the given attribute name to the attribute section.
        
        Parameters
        ----------
        name : `str`
        """
        self.relations.setdefault(name, None)
    
    def serialize(self):
        """
        Serializes the attribute section to html string parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        object_ = self.object
        title = self.title
        path = self.path
        prefix = get_anchor_prefix_for(path)
        
        tier = get_tier_for(path) + 1
        parent_path = get_parent_path_of(path)
        
        if prefix:
            prefixed_title = f'{prefix}-{anchor_escape(title)}'
        else:
            prefixed_title = anchor_escape(title)
        
        yield '<h'
        yield str(tier)
        yield '>'
        yield from anchor_for_serializer(prefixed_title)
        yield html_escape(title)
        yield '<div class="underline"></div></h'
        yield str(tier)
        yield '>'
        
        relations = self.relations
        attribute_names = sorted(relations, key = name_sort_key)
        
        for attribute_name in attribute_names:
            attribute_unit = relations[attribute_name]
            yield '<div class="with_anchor">'
            
            if prefix:
                prefixed_name = f'{prefix}-{anchor_escape(attribute_name)}'
            else:
                prefixed_name = anchor_escape(attribute_name)
            
            yield from anchor_for_serializer(prefixed_name)
            yield html_escape(attribute_name)
            
            if attribute_unit is None:
                yield '</div>'
                yield '<br>'
                continue
            
            attribute_content = attribute_unit.sections[0][1]
            maybe_head = attribute_content[0]
            if type(maybe_head) is GravedAttributeDescription:
                separator = maybe_head.separator
                if separator == '(':
                    yield_space = False
                else:
                    yield_space = True
                
                if yield_space:
                    yield ' '
                
                yield html_escape(separator)
                
                if yield_space:
                    yield ' '
                
                yield graved_to_escaped(maybe_head.content, object_, parent_path, create_relative_sectioned_link)
                yield '</div>'
                attribute_content = attribute_content[1:]
                if not attribute_content:
                    continue
            else:
                yield '</div>'
            
            yield '<div class="sub_section">'
            yield from sub_section_serializer(attribute_content, object_, parent_path, create_relative_sectioned_link)
            yield '</div>'
            continue
        
        extra = self.extra
        if extra is None:
            return
        
        extra_content = extra.sections[0][1]
        yield from sub_section_serializer(extra_content, object_, parent_path, create_relative_sectioned_link)
        return
    
    def structurize(self):
        """
        Creates a section structure for the represented attribute section.
        
        Returns
        -------
        structure : ``Structure``
        """
        path = self.path
        prefix = get_anchor_prefix_for(path)
        
        children = []
        
        for child_title in sorted(self.relations, key = name_sort_key):
            if prefix:
                child_prefixed_title = f'{prefix}-{anchor_escape(child_title)}'
            else:
                child_prefixed_title = anchor_escape(child_title)
            
            child = Structure(child_title, child_prefixed_title, None)
            children.append(child)
        
        if not children:
            children = None
        
        title = self.title
        if prefix:
            prefixed_title = f'{prefix}-{anchor_escape(title)}'
        else:
            prefixed_title = anchor_escape(title)
        
        return Structure(title, prefixed_title, children)

PARAMETER_NAME_RP = re.compile('(\*{0,2}[a-zA-Z_]+[a-zA-Z_0-9]*)(?: *\: *(.+)?)?')
PARAMETER_OPTIONALITY_RP = re.compile('(.*?)(?:,? *([Oo]ptional)(?:,? *\(?([Kk]eyword [Oo]nly)\)?)?)?')
PARAMETER_DEFAULT_START_RP = re.compile('(.*?) *= *(.*?)')

PARAMETER_SHIFT_NAME = 0
PARAMETER_SHIFT_DESCRIPTION = 1
PARAMETER_SHIFT_TYPE = 2
PARAMETER_SHIFT_OPTIONAL = 3
PARAMETER_SHIFT_KEYWORD_ONLY = 4
PARAMETER_SHIFT_DEFAULT = 5

PARAMETER_MASK_NAME = 1 << PARAMETER_SHIFT_NAME
PARAMETER_MASK_DESCRIPTION = 1 << PARAMETER_SHIFT_DESCRIPTION
PARAMETER_MASK_TYPE = 1 << PARAMETER_SHIFT_TYPE
PARAMETER_MASK_OPTIONAL = 1 << PARAMETER_SHIFT_OPTIONAL
PARAMETER_MASK_KEYWORD_ONLY = 1 << PARAMETER_SHIFT_KEYWORD_ONLY
PARAMETER_MASK_DEFAULT = 1 << PARAMETER_SHIFT_DEFAULT


class ParameterSubSection:
    """
    A ``ParameterSection``'s parameter row.
    
    Parameters
    ----------
    default : `None`, ``GravedDescription``
        The default value of the parameter.
    description : `None`, `list` of ``GravedDescription``
        Description of the parameter.
    keyword_only : `bool`
        Whether the parameter is keyword only.
    name : `str`
        The parameter's name.
    optional : `bool`
        Whether the parameter is optional.
    type : `None`, ``GravedDescription``
        The parameter's type.
    """
    __slots__ = ('default', 'description', 'keyword_only', 'name', 'optional', 'type')
    
    def __init__(self, header, description):
        self.description = description
        header_contents = header.content.copy()
        
        parameter_name_part = header_contents[0]
        if isinstance(parameter_name_part, str):
            match = PARAMETER_NAME_RP.fullmatch(parameter_name_part)
            if match is None:
                name = ''
            else:
                name, after_part = match.groups()
                if after_part is None:
                    del header_contents[0]
                else:
                    header_contents[0] = after_part
        else:
            # should not happen
            name = ''
            
        self.name = name
        
        if header_contents:
            parameter_optionality_part = header_contents[-1]
            if isinstance(parameter_optionality_part, str):
                match = PARAMETER_OPTIONALITY_RP.fullmatch(parameter_optionality_part)
                if match is None:
                    optional = False
                    keyword_only = False
                else:
                    before_part, optional_part, keyword_only_part = match.groups()
                    
                    if (before_part is not None) and before_part:
                        header_contents[-1] = before_part
                    else:
                        del header_contents[-1]
                    
                    if (optional_part is None):
                        optional = False
                    else:
                        optional = True
                    
                    if (keyword_only_part is None):
                        keyword_only = False
                    else:
                        keyword_only = True
            else:
                optional = False
                keyword_only = False
        else:
            optional = False
            keyword_only = False
        
        if name.startswith('**'):
            optional = True
            keyword_only = True
        
        elif name.startswith('*'):
            optional = True
            keyword_only = False
        
        self.optional = optional
        self.keyword_only = keyword_only
        
        if header_contents:
            type_ = header
            default = header.split_at(PARAMETER_DEFAULT_START_RP)
        else:
            type_ = None
            default = None
        
        self.type = type_
        self.default = default
    
    
    def get_mask(self):
        """
        Gets the mask of the used features by the parameter sub part.
        
        Returns
        -------
        mask : `int`
        """
        # We add name every time
        mask = PARAMETER_MASK_NAME
        
        if (self.description is not None):
            mask |= PARAMETER_MASK_DESCRIPTION
        
        if self.keyword_only:
            mask |= PARAMETER_MASK_KEYWORD_ONLY
        
        if self.optional:
            mask |= PARAMETER_MASK_OPTIONAL
        
        if (self.type is not None):
            mask |= PARAMETER_MASK_TYPE
        
        if (self.default is not None):
            mask |= PARAMETER_MASK_DEFAULT
        
        return mask
    
    
    def serialize(self, parent, mask):
        """
        Serializes the parameter sub section to html string parts.
        
        This method is a generator.
        
        Parameters
        ----------
        parent : ``ParameterSerializer``
            The parent parameter serialiser.
        mask : `int`
            Mask to serialise the mask based of.
        
        Yields
        ------
        html : `str`
        """
        yield '<tr>'
        
        if mask & PARAMETER_MASK_NAME:
            yield '<td class="parameter_table_name">'
            
            name = self.name
            if name:
                yield html_escape(name)
            
            yield '</td>'
        
        if mask & PARAMETER_MASK_TYPE:
            yield '<td class="parameter_table_type">'
            
            type_ = self.type
            if (type_ is not None):
                yield from description_serializer(type_, parent.object, parent.path, create_relative_sectioned_link)
            
            yield '</td>'
        
        if mask & PARAMETER_MASK_OPTIONAL:
            yield '<td class="parameter_table_optional">'
            
            if self.optional:
                yield YES_SVG
            
            yield '</td>'
        
        if mask & PARAMETER_MASK_KEYWORD_ONLY:
            yield '<td class="parameter_table_keyword_only">'
            
            if self.keyword_only:
                yield YES_SVG
            
            yield '</td>'
        
        if mask & PARAMETER_MASK_DEFAULT:
            yield '<td class="parameter_table_default">'
            
            default = self.default
            if (default is not None):
                yield from description_serializer(default, parent.object, parent.path, create_relative_sectioned_link)
            
            yield '</td>'
        
        if mask & PARAMETER_MASK_DESCRIPTION:
            yield '<td class="parameter_table_description">'
            
            description = self.description
            if (description is not None):
                yield from sub_section_serializer(description, parent.object, get_parent_path_of(parent.path),
                    create_relative_sectioned_link)
            
            yield '</td>'
        
        yield '</tr>'
    
    
    def __repr__(self):
        """Returns the parameter sub section's representation."""
        repr_parts = ['<', self.__class__.__name__, ' name=', repr(self.name)]
        
        type_ = self.type
        if (type_ is not None):
            repr_parts.append(', type=')
            repr_parts.append(reprlib.repr(type_))
        
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(reprlib.repr(description))
        
        if self.optional:
            repr_parts.append(', optional=True')
        
        if self.keyword_only:
            repr_parts.append(', keyword only=True')
        
        repr_parts.append('>')
        return ''.join(repr_parts)


class ParameterSection:
    """
    Represents a parameter's section.
    
    Attributes
    ----------
    parameter_sub_sections : `list` of `ParameterSubSection`
        The stored parameters.
    object : ``UnitBase``
        The owner unit.
    path : ``QualPath``
        Path to use instead of the objects'.
    """
    __slots__ = ('parameter_sub_sections', 'object', 'path')
    
    def __init__(self, section_content, object_, path):
        """
        Creates a new parameter section instance with the given initial content.
        
        Parameters
        ----------
        section_content : `list` of `Any`
            Contained section parts.
        object_ : ``TypeUnit``
            The owner type-unit.
        path : `None`, ``QualPath``
            Path to use instead of the objects'.
        """
        self.parameter_sub_sections = []
        self.object = object_
        if path is None:
            path = object_.path
        self.path = path
        self.extend(section_content)
    
    
    def extend(self, section_content):
        """
        Extends the parameter section with the given content.
        
        Parameters
        ----------
        section_content : `list` of `Any`
            Contained section parts.
        """
        header = None
        for section_content in section_content:
            # section content can be either ``GraveDescription``, `list` of it.
            
            if isinstance(section_content, GravedDescription):
                if header is None:
                    header = section_content
                else:
                    # Unlucky. Parameter definition without description
                    self.parameter_sub_sections.append(ParameterSubSection(header, None))
                    header = section_content
            
            else:
                if header is None:
                    # nani desuka?
                    pass
                else:
                    self.parameter_sub_sections.append(ParameterSubSection(header, section_content))
                    header = None
        
        if (header is not None):
            self.parameter_sub_sections.append(ParameterSubSection(header, None))
    
    
    def serialize(self):
        """
        Serializes the section to html string parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        mask = 0
        for parameter_sub_section in self.parameter_sub_sections:
            mask |= parameter_sub_section.get_mask()
        
        yield '<table class="parameter_table"><thead><tr>'
        
        if mask & PARAMETER_MASK_NAME:
            yield '<th class="parameter_table_name">Parameter</th>'
        
        if mask & PARAMETER_MASK_TYPE:
            yield '<th class="parameter_table_type">Type</th>'
        
        if mask & PARAMETER_MASK_OPTIONAL:
            yield '<th class="parameter_table_optional">Optional</th>'
        
        if mask & PARAMETER_MASK_KEYWORD_ONLY:
            yield '<th class="parameter_table_keyword_only">Keyword only</th>'
        
        if mask & PARAMETER_MASK_DEFAULT:
            yield '<th class="parameter_table_default">Default</th>'
        
        if mask & PARAMETER_MASK_DESCRIPTION:
            yield '<th class="parameter_table_description">Description</th>'
        
        yield '</tr></thead><tbody>'
        
        for parameter_sub_section in self.parameter_sub_sections:
            yield from parameter_sub_section.serialize(self, mask)
        
        yield '</tbody></table>'
    
    
    def structurize(self):
        """
        Returns the structure of the section.
        
        Returns
        -------
        structure : `None`, ``Structure``
            If the section is unnamed, returns `None`.
        """
        # Since we are building a big table this time, we have nothing to do with this.
        return None


SECTION_NAME_TYPE_DEFAULT_RELATIONS = {
    'Usage': SECTION_TYPE_USAGES,
    'Usages': SECTION_TYPE_USAGES,
    'Yields': SECTION_TYPE_YIELDS,
    'Returns': SECTION_TYPE_RETURNS,
    'Raises': SECTION_TYPE_RAISES,
    'Notes': SECTION_TYPE_NOTES,
}


class TypeSerializer:
    """
    Converts the given type docs to topically broken down parts.
    
    Attributes
    ----------
    object : ``TypeUnit``
        The represented type unit.
    sections : `list` of (``SimpleSection``, ``AttributeSection``, ``UnitSection``, ``ParameterSection``)
        The type's sections.
    path : ``QualPath``
        Path to use instead of the object's.
    """
    __slots__ = ('object', 'sections', 'path')
    
    def __init__(self, object_, path=None):
        """
        Parameters
        ----------
        object_ : ``UnitBase``
            The respective unit.
        path : `None`, ``QualPath`` = `None`, Optional
            Path to use instead of the object's.
        """
        self.object = object_
        if path is None:
            path = object_.path
        self.path = path
        
        section_parts = {}
        
        docs = object_.docs
        if docs is not None:
            for section_name, section_content in docs.sections:
                if section_name is None:
                    section_parts[SECTION_TYPE_DEFAULT] = SimpleSection(None, section_content, object_, path)
                    continue
                
                try:
                    section_type_value = SECTION_NAME_TYPE_DEFAULT_RELATIONS[section_name]
                except KeyError:
                    pass
                else:
                    section_parts[section_type_value] = SimpleSection(section_name, section_content, object_, path)
                    continue
                
                
                if section_name in ('Attributes', 'Instance Attributes'):
                    mentioned_names = get_attribute_section_mentioned_names(section_content)
                    section_parts[SECTION_TYPE_INSTANCE_ATTRIBUTES] = \
                        AttributeSection(section_name, mentioned_names, object_, None)
                    continue
                
                if section_name in ('Class Attributes', 'Type Attributes'):
                    mentioned_names = get_attribute_section_mentioned_names(section_content)
                    section_parts[SECTION_TYPE_OTHER_PARAMETERS] = \
                        AttributeSection(section_name, mentioned_names, object_, None)
                    continue
                
                if ATTRIBUTE_SECTION_NAME_RP.fullmatch(section_name) is not None:
                    try:
                        existing = section_parts[SECTION_TYPE_X_ATTRIBUTES]
                    except KeyError:
                        existing = section_parts[SECTION_TYPE_X_ATTRIBUTES] = []
                    
                    mentioned_names = get_attribute_section_mentioned_names(section_content)
                    existing.append(AttributeSection(section_name, mentioned_names, object_, path))
                    continue
                
                try:
                    existing = section_parts[SECTION_TYPE_X]
                except KeyError:
                    existing = section_parts[SECTION_TYPE_X] = []
                
                existing.append(SimpleSection(section_name, section_content, object_, path))
                continue
        
        collected_instance_attributes = []
        collected_properties = []
        collected_methods = []
        collected_classes = []
        
        # Do not mention `ClassAttributeUnit`. There might be a mention, why there are not mentioned. For example at the
        # case of precreation they are probably mentioned inside of a table.
        
        for item in object_.references.items():
            unit_type = type(item[1])
            if unit_type is PropertyUnit:
                container = collected_properties
            elif unit_type is InstanceAttributeUnit:
                container = collected_instance_attributes
            elif unit_type is FunctionUnit:
                container = collected_methods
            elif unit_type is TypeUnit:
                container = collected_classes
            else:
                # We do not care.
                continue
            
            container.append(item)
            continue
        
        for instance_attribute_name, _ in collected_instance_attributes:
            try:
                section_attributes_i = section_parts[SECTION_TYPE_INSTANCE_ATTRIBUTES]
            except KeyError:
                section_attributes_i = None
            else:
                if section_attributes_i.contains(instance_attribute_name):
                    continue
            
            # Extra?
            try:
                section_attributes_x = section_parts[SECTION_TYPE_X_ATTRIBUTES]
            except KeyError:
                pass
            else:
                for section_attribute_x in section_attributes_x:
                    if section_attribute_x.contains(instance_attribute_name):
                        found = True
                        break
                else:
                    found = False
                
                if found:
                    continue
            
            # Maybe misplaced?
            try:
                section_attributes_c = section_parts[SECTION_TYPE_CLASS_ATTRIBUTES]
            except:
                pass
            else:
                if section_attributes_c.contains(instance_attribute_name):
                    continue
            
            
            if section_attributes_i is None:
                section_parts[SECTION_TYPE_INSTANCE_ATTRIBUTES] = \
                    AttributeSection('Attributes', {instance_attribute_name}, object_, path)
            else:
                section_attributes_i.add(instance_attribute_name)
            continue
        
        if collected_properties:
            collected_properties = [item for item in sorted(collected_properties, key = item_sort_key)]
            section_parts[SECTION_TYPE_PROPERTIES] = UnitSection('Properties', object_, collected_properties, path)
        
        if collected_methods:
            collected_methods = [item for item in sorted(collected_methods, key = item_sort_key)]
            section_parts[SECTION_TYPE_METHODS] = UnitSection('Methods', object_, collected_methods, path)
        
        if collected_classes:
            collected_classes = [item for item in sorted(collected_classes, key = item_sort_key)]
            section_parts[SECTION_TYPE_CLASSES] = UnitSection('Classes', object_, collected_classes, path)
        
        sections = []
        
        for item in sorted(section_parts.items()):
            section = item[1]
            if type(section) is list:
                sections.extend(section)
            else:
                sections.append(section)
        
        self.sections = sections
        
    def serialize(self):
        """
        Serializes the attribute section to html string parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        path = self.path
        
        name = path.parts[-1]
        prefix = get_anchor_prefix_for(path)
        tier = get_tier_for(path)
        
        yield '<h'
        yield str(tier)
        yield '>'
        yield from anchor_for_serializer(anchor_escape(prefix))
        yield html_escape(name)
        yield '<div class="underline"></div></h'
        yield str(tier)
        yield '>'
        
        if tier > 1:
            yield '<div class="sub_unit">'
        
        for section in self.sections:
            yield from section.serialize()
        
        if tier > 1:
            yield '</div>'
        
        return

    def structurize(self):
        """
        Creates a section structure for the represented unit.
        
        Returns
        -------
        structure : ``Structure``
        """
        children = []
        for section in self.sections:
            child = section.structurize()
            if child is None:
                continue
            
            children.append(child)
        
        if not children:
            children = None
        
        path = self.path
        title = path.parts[-1]
        prefixed_title = get_anchor_prefix_for(path)
        
        return Structure(title, prefixed_title, children)

class UnitListerSection:
    """
    Serializer to list units inside of a module section.
    
    Attributes
    ----------
    title : `str`
        The section's title.
    object : ``TypeUnit``
        The represented parent object.
    units : `list` of ``UnitBase``
        The contained units.
    path : ``QualPath``
        Path to use instead of the object's.
    """
    __slots__ = ('title', 'object', 'units', 'path',)
    
    def __init__(self, title, object_, units, path):
        """
        Creates a new unit lister section.
        
        Parameters
        ----------
        title : `str`
            The section's title.
        object_ : ``TypeUnit``
            The represented parent object.
        units : `list` of ``UnitBase``
            The contained units.
        path : ``QualPath``
            Path to use instead of the object's.
        """
        self.title = title
        self.object = object_
        self.units = units
        if path is None:
            path = object_.path
        self.path = path
    
    def serialize(self):
        """
        Serializes the unit listing.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        title = self.title
        path = self.path
        prefix = get_anchor_prefix_for(path)
        if prefix:
            prefix = f'{prefix}-{anchor_escape(title)}'
        else:
            prefix = anchor_escape(title)
        
        tier = get_tier_for(path) + 1
        
        yield '<h'
        yield str(tier)
        yield '>'
        yield from anchor_for_serializer(prefix)
        yield html_escape(title)
        yield '<div class="underline"></div></h'
        yield str(tier)
        yield '><ul class="unit_listing">'
        
        for unit in self.units:
            url = create_relative_link(path, unit.path)
            
            name = html_escape(unit.name)
            yield '<li><a href="'
            yield url
            yield '" title="'
            yield name
            yield '">'
            yield name
            yield '</a></li>'
        
        yield '</ul>'
    
    def structurize(self):
        """
        Creates a section structure for the represented unit listing.
        
        Returns
        -------
        structure : ``Structure``
        """
        title = self.title
        child_prefixed_title = anchor_escape(title)
        return Structure(title, child_prefixed_title, None)


class ModuleSerializer:
    """
    Collects the given module to topically broken down parts.
    
    Attributes
    ----------
    object : ``ModuleUnit``
        The represented type unit.
    sections : `list` of (``SimpleSection``, ``AttributeSection``, ``UnitSection``)
        The type's sections.
    path : `None`, ``QualPath``
        Path to use instead of the object's.
    """
    __slots__ = ('object', 'sections', 'path')
    
    def __init__(self, object_, path=None):
        """
        Creates a module serializer to html serialize a module.
        
        Parameters
        ----------
        object_ : ``ModuleUnit``
            The module to serialize.
        path : `None`, ``QualPath``
            Path to use instead of the object's.
        """
        self.object = object_
        if path is None:
            path = object_.path
        self.path = path
        section_parts = {}
        
        docs = object_.docs
        if (docs is not None):
            for section_name, section_content in docs.sections:
                if section_name is None:
                    section_parts[SECTION_TYPE_DEFAULT] = SimpleSection(None, section_content, object_, path)
                    continue
                
                try:
                    section_type_value = SECTION_NAME_TYPE_DEFAULT_RELATIONS[section_name]
                except KeyError:
                    pass
                else:
                    section_parts[section_type_value] = SimpleSection(section_name, section_content, object_, path)
                    continue
                
                if section_name in ('Attributes', 'Instance Attributes'):
                    mentioned_names = get_attribute_section_mentioned_names(section_content)
                    section_parts[SECTION_TYPE_INSTANCE_ATTRIBUTES] = \
                        AttributeSection(section_name, mentioned_names, object_, None)
                    continue
                
                if section_name in ('Class Attributes', 'Type Attributes'):
                    mentioned_names = get_attribute_section_mentioned_names(section_content)
                    section_parts[SECTION_TYPE_OTHER_PARAMETERS] = \
                        AttributeSection(section_name, mentioned_names, object_, None)
                    continue
                
                if ATTRIBUTE_SECTION_NAME_RP.fullmatch(section_name) is not None:
                    try:
                        existing = section_parts[SECTION_TYPE_X_ATTRIBUTES]
                    except KeyError:
                        existing = section_parts[SECTION_TYPE_X_ATTRIBUTES] = []
                    
                    mentioned_names = get_attribute_section_mentioned_names(section_content)
                    existing.append(AttributeSection(section_name, mentioned_names, object_, None))
                    continue
                
                try:
                    existing = section_parts[SECTION_TYPE_X]
                except KeyError:
                    existing = section_parts[SECTION_TYPE_X] = []
                
                existing.append(SimpleSection(section_name, section_content, object_, path))
                continue
        
        collected_functions = []
        collected_classes = []
        collected_modules = []
        for item in object_.references.items():
            unit_type = type(item[1])
            if unit_type is FunctionUnit:
                container = collected_functions
            elif unit_type is TypeUnit:
                container = collected_classes
            elif unit_type is ModuleUnit:
                container = collected_modules
            else:
                continue
            
            container.append(item)
            continue
        
        if collected_functions:
            collected_functions = [item[1] for item in sorted(collected_functions, key = item_sort_key)]
            section_parts[SECTION_TYPE_FUNCTIONS] = UnitListerSection('Functions', object_, collected_functions, path)
        
        if collected_classes:
            collected_classes = [item[1] for item in sorted(collected_classes, key = item_sort_key)]
            section_parts[SECTION_TYPE_CLASSES] = UnitListerSection('Classes', object_, collected_classes, path)
        
        if collected_modules:
            collected_modules = [item[1] for item in sorted(collected_modules, key = item_sort_key)]
            section_parts[SECTION_TYPE_MODULES] = UnitListerSection('Modules', object_, collected_modules, path)
        
        sections = []
        
        for item in sorted(section_parts.items()):
            section = item[1]
            if type(section) is list:
                sections.extend(section)
            else:
                sections.append(section)
        
        self.sections = sections
    
    def serialize(self):
        """
        Serializes the attribute section to html string parts.
        
        This method is a generator.
        
        Yields
        ------
        html : `str`
        """
        path = self.path
        name = path.parts[-1]
        
        yield '<h1>'
        yield from anchor_for_serializer(anchor_escape(name))
        yield html_escape(name)
        yield '<div class="underline"></div></h1>'
        
        for section in self.sections:
            yield from section.serialize()
        
        return
    
    def structurize(self):
        """
        Creates a section structure for the represented unit.
        
        Returns
        -------
        structure : ``Structure``
        """
        children = []
        for section in self.sections:
            child = section.structurize()
            if child is None:
                continue
            
            children.append(child)
        
        if not children:
            children = None
        
        title = self.path.parts[-1]
        prefixed_title = anchor_escape(title)
        
        return Structure(title, prefixed_title, children)

UNIT_CONVERSION_TABLE = {
    PropertyUnit: FunctionOrPropertySerializer,
    FunctionUnit: FunctionOrPropertySerializer,
    TypeUnit: TypeSerializer,
    ModuleUnit: ModuleSerializer,
}

@export
def html_serialize_docs_extended(object_, get_html, get_structure):
    """
    Serializes the given docs to one big html code.
    
    Not like ``html_serialize_docs`` this serialization serializes all the methods and properties of the object as well.
    
    Parameters
    ----------
    object_ : ``UnitBase``
        The respective unit.
    get_html : `bool`
        Whether html code string should be created and returned.
    get_structure : `bool`
        Whether structure of the created html code string should be created.
    
    Returns
    -------
    html : `None`  or `str`
        The created html code if possible to create or applicable.
    structure : `None`, ``Structure``
        The structure of the html code if possible to create or applicable.
    """
    try:
        serializer_type = UNIT_CONVERSION_TABLE[type(object_)]
    except KeyError:
        html = None
        structure = None
    else:
        serializer = serializer_type(object_)
        if get_html:
            html = ''.join(serializer.serialize())
        else:
            html = None
        
        if get_structure:
            structure = serializer.structurize()
        else:
            structure = None
    
    return html, structure



from .command import *

from .call import *
from .constants import *
from .external import *
from .helpers import *
from .lookup import *
from .registration import *


__all__ = (
    *command.__all__,
    
    *call.__all__,
    *constants.__all__,
    *external.__all__,
    *helpers.__all__,
    *registration.__all__,
    *lookup.__all__,
)



__all__ = ('load_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.load)
def load_plugin(name, *, blocking=True, deep=True):
    return PLUGIN_LOADER.load(name, blocking=blocking, deep=deep)



__all__ = (
    'AutoModerationActionType', 'AutoModerationEventType', 'AutoModerationKeywordPresetType',
    'AutoModerationRuleTriggerType'
)

from scarletio import export

from ..bases import Preinstance as P, PreinstancedBase

from .action_metadata import AutoModerationActionMetadataSendAlertMessage, AutoModerationActionMetadataTimeout
from .trigger_metadata import (
    AutoModerationRuleTriggerMetadataKeywordPreset, AutoModerationRuleTriggerMetadataKeyword,
    AutoModerationRuleTriggerMetadataMentionSpam
)


class AutoModerationActionType(PreinstancedBase):
    """
    Represents an ``AutoModerationAction``'s type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation action type.
    name : `str`
        The default name of the auto moderation action type.
    metadata_type : `None`, ``AutoModerationActionMetadataBase``
        The action type's respective metadata type.
    
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AutoModerationActionType``) items
        Stores the predefined auto moderation action types. This container is accessed when translating a Discord side
        identifier of a auto moderation action type. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `type` = `str`
        The auto moderation action types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the auto moderation action types.
    
    Every predefined auto moderation action type is also stored as a class attribute:
    
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | Class attribute name  | Name                  | Value     | Metadata type                                     | Description                                                           |
    +=======================+=======================+===========+===================================================+=======================================================================+
    | none                  | none                  | 0         | `None`                                            | N/A                                                                   |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | block_message         | block message         | 1         | `None`                                            | Blocks the message's content according to the rule.                   |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | send_alert_message    | send alert message    | 2         | ``AutoModerationActionMetadataSendAlertMessage``  | Sends an alert message to the specified channel.                      |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | timeout               | timeout               | 3         | ``AutoModerationActionMetadataTimeout``           | Timeouts the user. Only applicable for `keyword` rules. Max 4 weeks.  |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    INSTANCES = {}
    VALUE_TYPE = int

    @classmethod
    def _from_value(cls, value):
        """
        Creates a new auto moderation action type with the given value.
        
        Parameters
        ----------
        value : `int`
            The auto moderation action type's identifier value.
        
        Returns
        -------
        self : ``AutoModerationActionType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.metadata_type = None
        
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates an ``AutoModerationActionType`` and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the auto moderation action type.
        name : `str`
            The default name of the auto moderation action type.
        max_per_guild : `int`
            The native name of the auto moderation action type.
        metadata_type : `None`, ``AutoModerationRuleTriggerMetadataBase``
            The action type's respective metadata type.
        """
        self.value = value
        self.name = name
        self.metadata_type = metadata_type
        
        self.INSTANCES[value] = self
    
    # predefined
    none = P(0, 'none', None)
    block_message = P(1, 'block message', None)
    send_alert_message = P(2, 'send alert message', AutoModerationActionMetadataSendAlertMessage)
    timeout = P(3, 'timeout', AutoModerationActionMetadataTimeout)


class AutoModerationRuleTriggerType(PreinstancedBase):
    """
    Represents an auto moderation rule's trigger type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation trigger type.
    name : `str`
        The default name of the auto moderation trigger type.
    max_per_guild : `int`
        The maximal amount of rules of this type per guild.
    metadata_type : `None`, ``AutoModerationRuleTriggerMetadataBase``
        The trigger type's respective metadata type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AutoModerationRuleTriggerType``) items
        Stores the predefined auto moderation trigger types. This container is accessed when translating a Discord side
        identifier of a auto moderation trigger type. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `type` = `str`
        The auto moderation trigger types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the auto moderation trigger types.
    
    Every predefined auto moderation trigger type is also stored as a class attribute:
    
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    | Class attribute name  | Name              | Value     | Max per guild | Metadata type                                       |
    +=======================+===================+===========+===============+=====================================================+
    | none                  | none              | 0         | 0             | `None`                                              |
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    | keyword               | keyword           | 1         | 3             | ``AutoModerationRuleTriggerMetadataKeyword``        |
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    | harmful_link          | harmful link      | 2         | 1             | `None`                                              |
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    | spam                  | spam              | 3         | 1             | `None`                                              |
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    | keyword_preset        | keyword preset    | 4         | 1             | ``AutoModerationRuleTriggerMetadataKeywordPreset``  |
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    | mention_spam          | mention spam      | 5         | 1             | ``AutoModerationRuleTriggerMetadataMentionSpam``    |
    +-----------------------+-------------------+-----------+---------------+-----------------------------------------------------+
    """
    __slots__ = ('max_per_guild', 'metadata_type')
    
    INSTANCES = {}
    VALUE_TYPE = int

    @classmethod
    def _from_value(cls, value):
        """
        Creates a new auto moderation trigger type with the given value.
        
        Parameters
        ----------
        value : `int`
            The auto moderation trigger type's identifier value.
        
        Returns
        -------
        self : ``AutoModerationRuleTriggerType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.max_per_guild = 1
        self.metadata_type = None
        
        return self
    
    
    def __init__(self, value, name, max_per_guild, metadata_type):
        """
        Creates an ``AutoModerationRuleTriggerType`` and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the auto moderation trigger type.
        name : `str`
            The default name of the auto moderation trigger type.
        max_per_guild : `int`
            The native name of the auto moderation trigger type.
        metadata_type : `None`, ``AutoModerationRuleTriggerMetadataBase``
            The trigger type's respective metadata type.
        """
        self.value = value
        self.name = name
        self.max_per_guild = max_per_guild
        self.metadata_type = metadata_type
        
        self.INSTANCES[value] = self
    
    # predefined
    none = P(0, 'none', 0, None)
    keyword = P(1, 'keyword', 3, AutoModerationRuleTriggerMetadataKeyword)
    harmful_link = P(2, 'harmful link', 1, None)
    spam = P(3, 'spam', 1, None)
    keyword_preset = P(4, 'keyword preset', 1, AutoModerationRuleTriggerMetadataKeywordPreset)
    mention_spam = P(5, 'mention spam', 1, AutoModerationRuleTriggerMetadataMentionSpam)


class AutoModerationEventType(PreinstancedBase):
    """
    Represents an auto moderation rule's event type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation event type.
    name : `str`
        The default name of the auto moderation event type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AutoModerationEventType``) items
        Stores the predefined auto moderation event types. This container is accessed when translating a Discord side
        identifier of a auto moderation event type. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `type` = `str`
        The auto moderation event types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the auto moderation event types.
    
    Every predefined auto moderation event type is also stored as a class attribute:
    
    +-----------------------+-------------------+-----------+
    | Class attribute name  | Name              | Value     |
    +=======================+===================+===========+
    | none                  | none              | 0         |
    +-----------------------+-------------------+-----------+
    | message_send          | message send      | 1         |
    +-----------------------+-------------------+-----------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    # predefined
    none = P(0, 'none')
    message_send = P(1, 'message send')


@export
class AutoModerationKeywordPresetType(PreinstancedBase):
    """
    Represents an auto moderation keyword preset type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation keyword preset type.
    name : `str`
        The default name of the auto moderation keyword preset type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AutoModerationKeywordPresetType``) items
        Stores the predefined auto moderation keyword preset types. This container is accessed when translating a
        Discord side identifier of a auto moderation keyword preset type. The identifier value is used as a key to
        get it's wrapper side representation.
    VALUE_TYPE : `type` = `str`
        The auto moderation keyword preset types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the auto moderation keyword preset types.
    
    Every predefined auto moderation keyword preset type is also stored as a class attribute:
    
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | Class attribute name  | Name                  | Value     | Description                               |
    +=======================+=======================+===========+===========================================+
    | none                  | none                  | 0         | N/A                                       |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | cursing               | cursing               | 1         | Swearing or cursing.                      |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | sexually_suggestive   | sexually suggestive   | 2         | Sexually explicit behavior or activity.   |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    | slur                  | slur                  | 3         | Personal insult or hate speech.           |
    +-----------------------+-----------------------+-----------+-------------------------------------------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    # predefined
    none = P(0, 'none')
    cursing = P(1, 'cursing')
    sexually_suggestive = P(2, 'sexually suggestive')
    slur = P(3, 'slur')



__all__ = ('ActivityMetadataCustom',)


from scarletio import copy_docs, include

from ...utils import DATETIME_FORMAT_CODE, DISCORD_EPOCH_START, datetime_to_millisecond_unix_time, millisecond_unix_time_to_datetime

from .base import ActivityMetadataBase


create_partial_emoji_data = include('create_partial_emoji_data')
create_partial_emoji_from_data = include('create_partial_emoji_from_data')


class ActivityMetadataCustom(ActivityMetadataBase):
    """
    Represents a Discord custom activity.
    
    Attributes
    ----------
    created_at : `None`, `datetime`
        When the status was created.
    emoji : `None`, ``Emoji``
        The emoji of the activity. If it has no emoji, then set as `None`.
    state : `None`, `str`
        The activity's text under it's emoji. Defaults to `None`.
    """
    __slots__ = ('created_at', 'emoji', 'state', )
    
    @copy_docs(ActivityMetadataBase.__new__)
    def __new__(cls, keyword_parameters):
        self = object.__new__(cls)
        self.created_at = None
        self.emoji = None
        self.state = None
        return self
    
    
    @copy_docs(ActivityMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # created_at
        created_at = self.created_at
        if (created_at is not None):
            repr_parts.append(' created_at=')
            repr_parts.append(format(created_at, DATETIME_FORMAT_CODE))
            
            field_added = True
        else:
            field_added = False
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' emoji=')
            repr_parts.append(repr(emoji))
        
        # state
        state = self.state
        if (state is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' state=')
            repr_parts.append(repr(state))
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityMetadataBase.__hash__)
    def __hash__(self):
        state = self.state
        emoji = self.emoji
        if (state is None):
            if (emoji is None):
                hash_ = 0
            else:
                hash_ = emoji.id
        else:
            hash_ = hash(state)
            if (emoji is not None):
                hash_ ^=emoji.id
        
        return hash_
    
    
    @copy_docs(ActivityMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # created_at
        if self.created_at != other.created_at:
            return False
        
        # emoji
        if self.emoji != other.emoji:
            return False
        
        # state
        if self.state != other.state:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ActivityMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self._update_attributes(data)
        return self
    
    
    @copy_docs(ActivityMetadataBase.to_data)
    def to_data(self, *, include_internals=False, user=False):
        data = {}
        
        if include_internals:
            data['name'] = 'Custom Status'
            
            # created_at
            created_at = self.created_at
            if (created_at is not None):
                data['created_at'] = datetime_to_millisecond_unix_time(created_at)
            
            # emoji
            emoji = self.emoji
            if (emoji is not None):
                data['emoji'] = create_partial_emoji_data(emoji)
            
            # state
            state = self.state
            if (state is not None):
                data['state'] = state
        
        return data
    
    
    @copy_docs(ActivityMetadataBase._update_attributes)
    def _update_attributes(self, data):
        self.state = data.get('state', None)
        
        emoji_data = data.get('emoji', None)
        if emoji_data is None:
            emoji = None
        else:
            emoji = create_partial_emoji_from_data(emoji_data)
        self.emoji = emoji
        
        created_at = data.get('created_at', None)
        if (created_at is not None):
            created_at = millisecond_unix_time_to_datetime(created_at)
        self.created_at = created_at
    
    
    @copy_docs(ActivityMetadataBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = {}
        
        state = data.get('state', None)
        if self.state != state:
            old_attributes['state'] = self.state
            self.state = state
        
        emoji_data = data.get('emoji', None)
        if emoji_data is None:
            emoji = None
        else:
            emoji = create_partial_emoji_from_data(emoji_data)
        
        if self.emoji != emoji:
            old_attributes['emoji'] = self.emoji
            self.emoji = emoji
        
        created_at = data.get('created_at', None)
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = millisecond_unix_time_to_datetime(created_at)
        if self.created_at != created_at:
            old_attributes['created_at'] = self.created_at
            self.created_at = created_at
        
        return old_attributes
    
    
    @property
    def name(self):
        """
        Returns the activity's display text.
        
        Returns
        -------
        name : `str`
        """
        state = self.state
        emoji = self.emoji
        if (state is None):
            if (emoji is None):
                name = ''
            else:
                name = emoji.as_emoji
        else:
            if (emoji is None):
                name = state
            else:
                name = f'{emoji.as_emoji} {state}'
        
        return name
